"""
This module provides functionality to load in JSON map files from
the Tiled Map Editor. This is achieved using the pytiled-parser
library.

For more info on Tiled see: https://www.mapeditor.org/
For more info on pytiled-parser see: https://github.com/Beefy-Swain/pytiled_parser
"""

from __future__ import annotations

import copy
import math
import os
from collections import OrderedDict
from collections.abc import Callable, Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import pytiled_parser
import pytiled_parser.tiled_object
from pytiled_parser import Color

import arcade
from arcade import (
    Sprite,
    SpriteList,
    TextureAnimation,
    TextureAnimationSprite,
    TextureKeyframe,
    get_window,
)
from arcade.hitbox import HitBoxAlgorithm, RotatableHitBox
from arcade.types import RGBA255
from arcade.types import Color as ArcadeColor

if TYPE_CHECKING:
    from arcade import DefaultTextureAtlas, Texture
    from arcade.texture_atlas import TextureAtlasBase

from pyglet.math import Vec2

from arcade.math import rotate_point
from arcade.resources import resolve
from arcade.types import Point2, TiledObject

_FLIPPED_HORIZONTALLY_FLAG = 0x80000000
_FLIPPED_VERTICALLY_FLAG = 0x40000000
_FLIPPED_DIAGONALLY_FLAG = 0x20000000

__all__ = ["TileMap", "load_tilemap"]

prop_to_float = cast(Callable[[pytiled_parser.Property], float], float)


def _get_image_info_from_tileset(tile: pytiled_parser.Tile) -> tuple[int, int, int, int]:
    image_x = 0
    image_y = 0

    if not tile.tileset:
        raise ValueError("Tileset was None. This should not happen")

    if tile.tileset.image is not None:
        margin = tile.tileset.margin or 0
        spacing = tile.tileset.spacing or 0
        row = tile.id // tile.tileset.columns
        image_y = margin + row * (tile.tileset.tile_height + spacing)
        col = tile.id % tile.tileset.columns
        image_x = margin + col * (tile.tileset.tile_width + spacing)

    if tile.tileset.image:
        width = tile.tileset.tile_width
        height = tile.tileset.tile_height
    else:
        image_x = tile.x
        image_y = tile.y
        width = tile.width
        height = tile.height

    return image_x, image_y, width, height


def _get_image_source(
    tile: pytiled_parser.Tile,
    map_directory: str | None,
) -> Path | None:
    image_file = None
    if tile.image:
        image_file = tile.image
    elif tile.tileset is not None:
        image_file = tile.tileset.image

    if not image_file:
        print(
            f"Warning for tile {tile.id}, no image source listed either for "
            "individual tile, or as a tileset."
        )
        return None

    if os.path.exists(image_file):
        return image_file

    if map_directory:
        try2 = Path(map_directory, image_file)
        if os.path.exists(try2):
            return try2

    print(f"Warning, can't find image {image_file} for tile {tile.id}")
    return None


def _may_be_flip(tile: pytiled_parser.Tile, texture: Texture) -> Texture:
    if tile.flipped_diagonally:
        texture = texture.flip_diagonally()
    if tile.flipped_horizontally:
        texture = texture.flip_horizontally()
    if tile.flipped_vertically:
        texture = texture.flip_vertically()
    return texture


class TileMap:
    """
    Class that represents a fully parsed and loaded map from Tiled.
    For examples on how to use this class, see: :ref:`platformer_part_twelve`

    Args:
        map_file:
            A JSON map file for a Tiled map to initialize from
        scaling:
            Global scaling to apply to all Sprites.
        layer_options:
            Extra parameters for each layer.
        use_spatial_hash:
            If set to True, this will make moving a sprite
            in the SpriteList slower, but it will speed up collision detection
            with items in the SpriteList. Great for doing collision detection
            with static walls/platforms.
        hit_box_algorithm:
            The hit box algorithm to use for the Sprite's in this layer.
        tiled_map:
            An already parsed pytiled-parser map object.
            Passing this means that the ``map_file`` argument will be ignored, and the pre-parsed
            map will instead be used. This can be helpful for working with Tiled World files.
        offset:
            Can be used to offset the position of all sprites and objects
            within the map. This will be applied in addition to any offsets from Tiled. This value
            can be overridden with the layer_options dict.
        texture_atlas:
            A default texture atlas to use for the SpriteLists created by this map.
            If not supplied the global default atlas will be used.
        lazy:
            SpriteLists will be created lazily.
        texture_cache_manager:
            The texture cache manager to use for loading textures.

    The ``layer_options`` parameter can be used to specify per layer arguments.
    The available options for this are:

    - ``use_spatial_hash`` - A boolean to enable spatial hashing on this layer's SpriteList.
    - ``scaling`` - A float providing layer specific Sprite scaling.
    - ``hit_box_algorithm`` - The hit box algorithm to use for the Sprite's in this layer.
    - ``offset`` - A tuple containing X and Y position offsets for the layer
    - ``custom_class`` - All objects in the layer are created from this class instead of Sprite. \
                       Must be subclass of Sprite.
    - ``custom_class_args`` - Custom arguments, passed into the constructor of the custom_class
    - ``texture_atlas`` - A texture atlas to use for the SpriteList from this layer, if none is \
        supplied then the one defined at the map level will be used.

        Example configuring layer options for a layer named "Platforms"::

            layer_options = {
                "Platforms": {
                    "use_spatial_hash": True,
                    "scaling": 2.5,
                    "offset": (-128, 64),
                    "custom_class": Platform,
                    "custom_class_args": {
                        "health": 100
                    }
                },
            }

    The keys and their values in each layer are passed to the layer processing functions
    using the `**` operator on the dictionary.
    """

    tiled_map: pytiled_parser.TiledMap
    """
    The pytiled-parser map object. This can be useful for implementing features
    that aren't supported by this class by accessing the raw map data directly.
    """

    width: float
    "The width of the map in tiles. This is the number of tiles, not pixels."

    height: float
    "The height of the map in tiles. This is the number of tiles, not pixels."

    tile_width: float
    "The width in pixels of each tile."

    tile_height: float
    "The height in pixels of each tile."

    background_color: Color | None
    "The background color of the map."

    scaling: float
    "A global scaling value to be applied to all Sprites in the map."

    sprite_lists: dict[str, SpriteList]
    """A dictionary mapping SpriteLists to their layer names. This is used
                    for all tile layers of the map."""

    object_lists: dict[str, list[TiledObject]]
    """
    A dictionary mapping TiledObjects to their layer names. This is used
    for all object layers of the map.
    """

    offset: Vec2
    "A tuple containing the X and Y position offset values."

    def __init__(
        self,
        map_file: str | Path = "",
        scaling: float = 1.0,
        layer_options: dict[str, dict[str, Any]] | None = None,
        use_spatial_hash: bool = False,
        hit_box_algorithm: HitBoxAlgorithm | None = None,
        tiled_map: pytiled_parser.TiledMap | None = None,
        offset: Vec2 = Vec2(0, 0),
        texture_atlas: TextureAtlasBase | None = None,
        lazy: bool = False,
        texture_cache_manager: arcade.TextureCacheManager | None = None,
    ) -> None:
        if not map_file and not tiled_map:
            raise AttributeError(
                "Initialized TileMap with an empty map_file or no tiled_map argument"
            )

        if tiled_map:
            self.tiled_map = tiled_map
        else:
            # If we should pull from local resources, replace with proper path
            map_file = resolve(map_file)

            # This attribute stores the pytiled-parser map object
            self.tiled_map = pytiled_parser.parse_map(map_file)

        if self.tiled_map.infinite:
            raise AttributeError(
                "Attempted to load an infinite TileMap. Arcade currently cannot load "
                "infinite maps. Disable the infinite map property and re-save the file."
            )

        if not texture_atlas:
            try:
                texture_atlas = get_window().ctx.default_atlas
            except RuntimeError:
                pass

        self._lazy = lazy
        self.texture_cache_manager = texture_cache_manager or arcade.texture.default_texture_cache

        # Set Map Attributes
        self.width = self.tiled_map.map_size.width
        self.height = self.tiled_map.map_size.height
        self.tile_width = self.tiled_map.tile_size.width
        self.tile_height = self.tiled_map.tile_size.height
        self.background_color = self.tiled_map.background_color

        # Global Layer Defaults
        self.scaling = scaling
        self.use_spatial_hash = use_spatial_hash
        self.hit_box_algorithm = hit_box_algorithm
        self.offset = offset

        # Dictionaries to store the SpriteLists for processed layers
        self.sprite_lists: dict[str, SpriteList] = OrderedDict()
        self.object_lists: dict[str, list[TiledObject]] = OrderedDict()
        self.properties = self.tiled_map.properties

        global_options = {  # type: ignore
            "scaling": self.scaling,
            "use_spatial_hash": self.use_spatial_hash,
            "hit_box_algorithm": self.hit_box_algorithm,
            "offset": self.offset,
            "custom_class": None,
            "custom_class_args": {},
            "texture_atlas": texture_atlas,
        }

        for layer in self.tiled_map.layers:
            if (layer.name in self.sprite_lists) or (layer.name in self.object_lists):
                raise AttributeError(
                    f"You have a duplicate layer name '{layer.name}' in your Tiled map. "
                    "Please use unique names for all layers and tilesets in your map."
                )
            self._process_layer(layer, global_options, layer_options)

    def _process_layer(
        self,
        layer: pytiled_parser.Layer,
        global_options: dict[str, Any],
        layer_options: dict[str, dict[str, Any]] | None = None,
    ) -> None:
        processed: SpriteList | tuple[SpriteList | None, list[TiledObject] | None]

        options = global_options

        if layer_options:
            if layer.name in layer_options:
                new_options = {
                    key: layer_options[layer.name].get(key, global_options[key])
                    for key in global_options
                }
                options = new_options

        if isinstance(layer, pytiled_parser.TileLayer):
            processed = self._process_tile_layer(layer, **options)
            self.sprite_lists[layer.name] = processed
        elif isinstance(layer, pytiled_parser.ObjectLayer):
            processed = self._process_object_layer(layer, **options)
            if processed[0]:
                sprite_list = processed[0]
                if sprite_list:
                    self.sprite_lists[layer.name] = sprite_list
            if processed[1]:
                object_list = processed[1]
                if object_list:
                    self.object_lists[layer.name] = object_list
        elif isinstance(layer, pytiled_parser.ImageLayer):
            processed = self._process_image_layer(layer, **options)
            self.sprite_lists[layer.name] = processed
        elif isinstance(layer, pytiled_parser.LayerGroup):
            if layer.layers:
                for sub_layer in layer.layers:
                    self._process_layer(sub_layer, global_options, layer_options)

    def get_cartesian(
        self,
        x: float,
        y: float,
    ) -> tuple[float, float]:
        """
        Given a set of coordinates in pixel units, this returns the cartesian coordinates.

        This assumes the supplied coordinates are pixel coordinates, and bases the cartesian
        grid off of the Map's tile size.

        If you have a map with 128x128 pixel Tiles, and you supply coordinates 500, 250 to
        this function you'll receive back 3, 2

        Args:
            x: The X Coordinate to convert
            y: The Y Coordinate to convert
        """
        x = math.floor(x / (self.tile_width * self.scaling))
        y = math.floor(y / (self.tile_height * self.scaling))

        return x, y

    def get_tilemap_layer(self, layer_path: str) -> pytiled_parser.Layer | None:
        """
        Given a path to a layer, this will attempt to return the layer.

        Args:
            layer_path:
                A string representing the path to the layer. For example,
                "Layer Group 1/Layer Group 2/Tile Layer 1"
        """
        assert isinstance(layer_path, str)

        def _get_tilemap_layer(my_path: list[str], layers):
            layer_name = my_path.pop(0)
            for my_layer in layers:
                if my_layer.name == layer_name:
                    if isinstance(my_layer, pytiled_parser.LayerGroup) and len(my_path) != 0:
                        return _get_tilemap_layer(my_path, my_layer.layers)
                    else:
                        return my_layer
            return None

        path = layer_path.strip("/").split("/")
        layer = _get_tilemap_layer(path, self.tiled_map.layers)
        return layer

    def _get_tile_by_gid(self, tile_gid: int) -> pytiled_parser.Tile | None:
        tile_ref: pytiled_parser.Tile | None

        flipped_diagonally = False
        flipped_horizontally = False
        flipped_vertically = False

        if tile_gid & _FLIPPED_HORIZONTALLY_FLAG:
            flipped_horizontally = True
            tile_gid -= _FLIPPED_HORIZONTALLY_FLAG

        if tile_gid & _FLIPPED_DIAGONALLY_FLAG:
            flipped_diagonally = True
            tile_gid -= _FLIPPED_DIAGONALLY_FLAG

        if tile_gid & _FLIPPED_VERTICALLY_FLAG:
            flipped_vertically = True
            tile_gid -= _FLIPPED_VERTICALLY_FLAG

        for tileset_key, tileset in self.tiled_map.tilesets.items():
            if tile_gid < tileset_key:
                continue

            # No specific tile info, but there is a tile sheet
            # print(
            #     f"data {tileset_key} {tileset.tiles} {tileset.image} "
            #     f"{tileset_key} {tile_gid} {tileset.tile_count}"
            # )
            if (
                tileset.image is not None
                and tileset_key <= tile_gid < tileset_key + tileset.tile_count
            ):
                tile_id = tile_gid - tileset_key
                existing_ref = None
                if tileset.tiles is not None:
                    if (tile_gid - tileset_key) in tileset.tiles:
                        existing_ref = tileset.tiles[tile_id]
                        existing_ref.image = tileset.image

                # No specific tile info, but there is a tile sheet
                if existing_ref:
                    tile_ref = existing_ref
                else:
                    tile_ref = pytiled_parser.Tile(id=tile_id, image=tileset.image)
            elif tileset.tiles is None and tileset.image is not None:
                # Not in this tileset, move to the next
                continue
            else:
                if tileset.tiles is None:
                    return None
                tile_ref = tileset.tiles.get(tile_gid - tileset_key)

            if tile_ref:
                my_tile = copy.copy(tile_ref)
                my_tile.tileset = tileset
                my_tile.flipped_vertically = flipped_vertically
                my_tile.flipped_diagonally = flipped_diagonally
                my_tile.flipped_horizontally = flipped_horizontally
                return my_tile

        print(f"Returning NO tile for {tile_gid}.")
        return None

    def _get_tile_by_id(
        self, tileset: pytiled_parser.Tileset, tile_id: int
    ) -> pytiled_parser.Tile | None:
        for tileset_key, cur_tileset in self.tiled_map.tilesets.items():
            if cur_tileset is tileset:
                if cur_tileset.tiles:
                    for tile_key, tile in cur_tileset.tiles.items():
                        if tile_id == tile.id:
                            return tile

        return None

    def _create_sprite_from_tile(
        self,
        tile: pytiled_parser.Tile,
        scaling: float = 1.0,
        hit_box_algorithm: HitBoxAlgorithm | None = None,
        custom_class: type | None = None,
        custom_class_args: dict[str, Any] = {},
    ) -> Sprite:
        """Given a tile from the parser, try and create a Sprite from it."""

        # --- Step 1, Find a reference to an image this is going to be based off of
        map_source = self.tiled_map.map_file
        map_directory = os.path.dirname(map_source)
        image_file = _get_image_source(tile, map_directory)

        if tile.animation:
            if not custom_class:
                custom_class = TextureAnimationSprite
            elif not issubclass(custom_class, TextureAnimationSprite):
                raise RuntimeError(
                    f"""
                    Tried to use a custom class {custom_class.__name__} for animated tiles
                    that doesn't subclass TextureAnimationSprite.
                    Custom classes for animated tiles must subclass TextureAnimationSprite.
                    """
                )
            # print(custom_class.__name__)
            args = {"path_or_texture": image_file, "scale": scaling}
            my_sprite = custom_class(**custom_class_args, **args)  # type: ignore
        else:
            if not custom_class:
                custom_class = Sprite
            elif not issubclass(custom_class, Sprite):
                raise RuntimeError(
                    f"""
                    Tried to use a custom class {custom_class.__name__} for
                    a tile that doesn't subclass arcade.Sprite.
                    Custom classes for tiles must subclass arcade.Sprite.
                    """
                )

            # Can image_file be None?
            image_x, image_y, width, height = _get_image_info_from_tileset(tile)
            texture = self.texture_cache_manager.load_or_get_texture(
                image_file,  # type: ignore
                x=image_x,
                y=image_y,
                width=width,
                height=height,
                hit_box_algorithm=hit_box_algorithm,
            )
            texture = _may_be_flip(tile, texture)

            args = {
                "path_or_texture": texture,  # type: ignore
                "scale": scaling,
            }
            my_sprite = custom_class(**custom_class_args, **args)  # type: ignore

        if tile.properties is not None and len(tile.properties) > 0:
            for key, value in tile.properties.items():
                my_sprite.properties[key] = value

        if tile.class_:
            my_sprite.properties["class"] = tile.class_

        # Add tile ID to sprite properties
        my_sprite.properties["tile_id"] = tile.id

        if tile.objects is not None:
            if not isinstance(tile.objects, pytiled_parser.ObjectLayer):
                print("Warning, tile.objects is not an ObjectLayer as expected.")
                return my_sprite

            if len(tile.objects.tiled_objects) > 1:
                if tile.image:
                    print(f"Warning, only one hit box supported for tile with image {tile.image}.")
                else:
                    print("Warning, only one hit box supported for tile.")

            for hitbox in tile.objects.tiled_objects:
                points: list[Point2] = []
                if isinstance(hitbox, pytiled_parser.tiled_object.Rectangle):
                    if hitbox.size is None:
                        print(
                            "Warning: Rectangle hitbox created for without a "
                            "height or width Ignoring."
                        )
                        continue

                    sx = hitbox.coordinates.x - (my_sprite.width / (scaling * 2))
                    sy = -(hitbox.coordinates.y - (my_sprite.height / (scaling * 2)))
                    ex = (hitbox.coordinates.x + hitbox.size.width) - (
                        my_sprite.width / (scaling * 2)
                    )
                    # issue #1068
                    # fixed size of rectangular hitbox
                    ey = -(hitbox.coordinates.y + hitbox.size.height) + (
                        my_sprite.height / (scaling * 2)
                    )

                    points = [(sx, sy), (ex, sy), (ex, ey), (sx, ey)]
                elif isinstance(hitbox, pytiled_parser.tiled_object.Polygon) or isinstance(
                    hitbox, pytiled_parser.tiled_object.Polyline
                ):
                    for point in hitbox.points:
                        adj_x = point.x + hitbox.coordinates.x - my_sprite.width / (scaling * 2)
                        adj_y = -(point.y + hitbox.coordinates.y - my_sprite.height / (scaling * 2))
                        adj_point = adj_x, adj_y
                        points.append(adj_point)

                    if points[0][0] == points[-1][0] and points[0][1] == points[-1][1]:
                        points.pop()
                elif isinstance(hitbox, pytiled_parser.tiled_object.Ellipse):
                    if not hitbox.size:
                        print(
                            f"Warning: Ellipse hitbox created without a height "
                            f" or width for {tile.image}. Ignoring."
                        )
                        continue

                    hw = hitbox.size.width / 2
                    hh = hitbox.size.height / 2
                    cx = hitbox.coordinates.x + hw
                    cy = hitbox.coordinates.y + hh

                    acx = cx - (my_sprite.width / (scaling * 2))
                    acy = cy - (my_sprite.height / (scaling * 2))

                    total_steps = 8
                    angles = [step / total_steps * 2 * math.pi for step in range(total_steps)]
                    for angle in angles:
                        x = hw * math.cos(angle) + acx
                        y = -(hh * math.sin(angle) + acy)
                        points.append((x, y))
                else:
                    print(f"Warning: Hitbox type {type(hitbox)} not supported.")

                if tile.flipped_vertically:
                    points = [(point[0], -point[1]) for point in points]

                if tile.flipped_horizontally:
                    points = [(-point[0], point[1]) for point in points]

                if tile.flipped_diagonally:
                    points = [(point[1], point[0]) for point in points]

                my_sprite.hit_box = RotatableHitBox(
                    cast(list[Point2], points),
                    position=my_sprite.position,
                    angle=my_sprite.angle,
                    scale=my_sprite.scale,
                )

        if tile.animation:
            key_frame_list = []
            for frame in tile.animation:
                frame_tile = self._get_tile_by_gid(tile.tileset.firstgid + frame.tile_id)  # type: ignore
                if frame_tile:
                    image_file = _get_image_source(frame_tile, map_directory)

                    if not frame_tile.tileset.image and image_file:  # type: ignore
                        texture = self.texture_cache_manager.load_or_get_texture(
                            image_file, hit_box_algorithm=hit_box_algorithm
                        )
                    elif image_file:
                        # No image for tile, pull from tilesheet
                        (
                            image_x,
                            image_y,
                            width,
                            height,
                        ) = _get_image_info_from_tileset(frame_tile)

                        texture = self.texture_cache_manager.load_or_get_texture(
                            image_file,
                            x=image_x,
                            y=image_y,
                            width=width,
                            height=height,
                            hit_box_algorithm=hit_box_algorithm,
                        )
                    else:
                        raise RuntimeError(
                            f"Warning: failed to load image for animation frame for "
                            f"tile '{frame_tile.id}', '{image_file}'."
                        )

                    texture = _may_be_flip(tile, texture)

                    key_frame = TextureKeyframe(  # type: ignore
                        texture=texture, duration=frame.duration, tile_id=frame.tile_id
                    )
                    key_frame_list.append(key_frame)

                    if len(key_frame_list) == 1:
                        my_sprite.texture = key_frame.texture

            # type: ignore
            cast(TextureAnimationSprite, my_sprite).animation = TextureAnimation(
                keyframes=key_frame_list
            )

        return my_sprite

    def _process_image_layer(
        self,
        layer: pytiled_parser.ImageLayer,
        texture_atlas: DefaultTextureAtlas,
        scaling: float = 1.0,
        use_spatial_hash: bool = False,
        hit_box_algorithm: HitBoxAlgorithm | None = None,
        offset: Vec2 = Vec2(0, 0),
        custom_class: type | None = None,
        custom_class_args: dict[str, Any] = {},
    ) -> SpriteList:
        sprite_list: SpriteList = SpriteList(
            use_spatial_hash=use_spatial_hash,
            atlas=texture_atlas,
            lazy=self._lazy,
        )

        map_source = self.tiled_map.map_file
        map_directory = os.path.dirname(map_source)
        image_file = layer.image

        if not os.path.exists(image_file) and (map_directory):
            try2 = Path(map_directory, image_file)
            if not os.path.exists(try2):
                print(f"Warning, can't find image {image_file} for Image Layer {layer.name}")
            image_file = try2

        my_texture = self.texture_cache_manager.load_or_get_texture(
            image_file,
            hit_box_algorithm=hit_box_algorithm,
        )

        if layer.transparent_color:
            # The pillow source doesn't annotate a return type for this method, but:
            # 1. The docstring does specify the returned object is sequence-like
            # 2. We convert to RGBA mode implicitly in load_or_get_texture above
            data: Sequence[RGBA255] = my_texture.image.getdata()  # type:ignore

            target = layer.transparent_color
            new_data = []
            for item in data:
                if item[0] == target[0] and item[1] == target[1] and item[2] == target[2]:
                    new_data.append((255, 255, 255, 0))
                else:
                    new_data.append(item)

            my_texture.image.putdata(new_data)

        if not custom_class:
            custom_class = Sprite
        elif not issubclass(custom_class, Sprite):
            raise RuntimeError(
                f"""
                    Tried to use a custom class {custom_class.__name__} for an
                    Image Layer that doesn't subclass arcade.Sprite.
                    Custom classes for image layers must subclass arcade.Sprite.
                """
            )

        args = {
            "filename": image_file,
            "scale": scaling,
            "path_or_texture": my_texture,
            "hit_box_algorithm": hit_box_algorithm,
        }

        my_sprite = custom_class(**custom_class_args, **args)  #  type: ignore

        if layer.properties:
            sprite_list.properties = layer.properties
            for key, value in layer.properties.items():
                my_sprite.properties[key] = value

        if layer.tint_color:
            my_sprite.color = ArcadeColor.from_iterable(layer.tint_color)

        if layer.opacity:
            my_sprite.alpha = int(layer.opacity * 255)

        my_sprite.center_x = ((layer.offset[0] * scaling) + my_sprite.width / 2) + offset[0]
        my_sprite.top = (
            self.tiled_map.map_size.height * self.tiled_map.tile_size[1] - layer.offset[1]
        ) * scaling + offset[1]

        sprite_list.visible = layer.visible
        sprite_list.append(my_sprite)
        return sprite_list

    def _process_tile_layer(
        self,
        layer: pytiled_parser.TileLayer,
        texture_atlas: DefaultTextureAtlas,
        scaling: float = 1.0,
        use_spatial_hash: bool = False,
        hit_box_algorithm: HitBoxAlgorithm | None = None,
        offset: Vec2 = Vec2(0, 0),
        custom_class: type | None = None,
        custom_class_args: dict[str, Any] = {},
    ) -> SpriteList:
        sprite_list: SpriteList = SpriteList(
            use_spatial_hash=use_spatial_hash,
            atlas=texture_atlas,
            lazy=self._lazy,
        )
        map_array = layer.data
        if TYPE_CHECKING:
            # Can never be None because we already detect and reject infinite maps
            assert map_array

        # Loop through the layer and add in the list
        for row_index, row in enumerate(map_array):
            for column_index, item in enumerate(row):
                # Check for an empty tile
                if item == 0:
                    continue

                tile = self._get_tile_by_gid(item)
                if tile is None:
                    raise ValueError(
                        f"Couldn't find tile for item {item} in layer "
                        f"'{layer.name}' in file '{self.tiled_map.map_file}'"
                        f"at ({column_index}, {row_index})."
                    )

                my_sprite = self._create_sprite_from_tile(
                    tile,
                    scaling=scaling,
                    hit_box_algorithm=hit_box_algorithm,
                    custom_class=custom_class,
                    custom_class_args=custom_class_args,
                )

                if my_sprite is None:
                    print(
                        f"Warning: Could not create sprite number {item} "
                        f"in layer '{layer.name}' {tile.image}"
                    )
                else:
                    my_sprite.center_x = (
                        column_index * (self.tiled_map.tile_size[0] * scaling) + my_sprite.width / 2
                    ) + offset[0]
                    my_sprite.center_y = (
                        (self.tiled_map.map_size.height - row_index - 1)
                        * (self.tiled_map.tile_size[1] * scaling)
                        + my_sprite.height / 2
                    ) + offset[1]

                    # Tint
                    if layer.tint_color:
                        my_sprite.color = ArcadeColor.from_iterable(layer.tint_color)

                    # Opacity
                    opacity = layer.opacity
                    if opacity:
                        my_sprite.alpha = int(opacity * 255)

                    sprite_list.visible = layer.visible
                    sprite_list.append(my_sprite)

                if layer.properties:
                    sprite_list.properties = layer.properties

        return sprite_list

    def _process_object_layer(
        self,
        layer: pytiled_parser.ObjectLayer,
        texture_atlas: DefaultTextureAtlas,
        scaling: float = 1.0,
        use_spatial_hash: bool = False,
        hit_box_algorithm: HitBoxAlgorithm | None = None,
        offset: Vec2 = Vec2(0, 0),
        custom_class: type | None = None,
        custom_class_args: dict[str, Any] = {},
    ) -> tuple[SpriteList | None, list[TiledObject] | None]:
        if not scaling:
            scaling = self.scaling

        sprite_list: SpriteList | None = None
        objects_list: list[TiledObject] | None = []

        shape: list[Point2] | tuple[int, int, int, int] | Point2 | None = None

        for cur_object in layer.tiled_objects:
            # shape: Optional[Point | PointList | Rect] = None
            if isinstance(cur_object, pytiled_parser.tiled_object.Tile):
                if not sprite_list:
                    sprite_list = SpriteList(
                        use_spatial_hash=use_spatial_hash,
                        atlas=texture_atlas,
                        lazy=self._lazy,
                    )

                tile = self._get_tile_by_gid(cur_object.gid)
                if tile is None:
                    raise Exception(f"Tile with gid not found: {cur_object.gid}")
                my_sprite = self._create_sprite_from_tile(
                    tile,
                    scaling=scaling,
                    hit_box_algorithm=hit_box_algorithm,
                    custom_class=custom_class,
                    custom_class_args=custom_class_args,
                )

                x = (cur_object.coordinates.x * scaling) + offset[0]
                y = (
                    (
                        self.tiled_map.map_size.height * self.tiled_map.tile_size[1]
                        - cur_object.coordinates.y
                    )
                    * scaling
                ) + offset[1]

                my_sprite.width = width = cur_object.size[0] * scaling
                my_sprite.height = height = cur_object.size[1] * scaling
                # center_x = width / 2
                # center_y = height / 2
                if cur_object.rotation:
                    rotation = math.radians(cur_object.rotation)
                else:
                    rotation = 0

                angle_degrees = math.degrees(rotation)
                rotated_center_x, rotated_center_y = rotate_point(
                    width / 2, height / 2, 0, 0, angle_degrees
                )

                my_sprite.position = (x + rotated_center_x, y + rotated_center_y)
                my_sprite.angle = angle_degrees

                if layer.tint_color:
                    my_sprite.color = ArcadeColor.from_iterable(layer.tint_color)

                opacity = layer.opacity
                if opacity:
                    my_sprite.alpha = int(opacity * 255)

                if cur_object.properties and "change_x" in cur_object.properties:
                    my_sprite.change_x = prop_to_float(cur_object.properties["change_x"])

                if cur_object.properties and "change_y" in cur_object.properties:
                    my_sprite.change_y = prop_to_float(cur_object.properties["change_y"])

                if cur_object.properties and "boundary_bottom" in cur_object.properties:
                    my_sprite.boundary_bottom = prop_to_float(
                        cur_object.properties["boundary_bottom"]
                    )

                if cur_object.properties and "boundary_top" in cur_object.properties:
                    my_sprite.boundary_top = prop_to_float(cur_object.properties["boundary_top"])

                if cur_object.properties and "boundary_left" in cur_object.properties:
                    my_sprite.boundary_left = prop_to_float(cur_object.properties["boundary_left"])

                if cur_object.properties and "boundary_right" in cur_object.properties:
                    my_sprite.boundary_right = prop_to_float(
                        cur_object.properties["boundary_right"]
                    )

                if cur_object.properties:
                    my_sprite.properties.update(cur_object.properties)

                if cur_object.class_:
                    my_sprite.properties["class"] = cur_object.class_

                if cur_object.name:
                    my_sprite.properties["name"] = cur_object.name

                sprite_list.visible = layer.visible
                sprite_list.append(my_sprite)
                continue
            elif isinstance(cur_object, pytiled_parser.tiled_object.Point):
                x = cur_object.coordinates.x * scaling
                y = (
                    self.tiled_map.map_size.height * self.tiled_map.tile_size[1]
                    - cur_object.coordinates.y
                ) * scaling

                shape = (x + offset[0], y + offset[1])
            elif isinstance(cur_object, pytiled_parser.tiled_object.Rectangle):
                if cur_object.size.width == 0 and cur_object.size.height == 0:
                    print(
                        f"WARNING: Tiled object with ID {cur_object.id} is a rectangle "
                        "with a width and height of 0. Loading it as a single point."
                    )
                    x = cur_object.coordinates.x * scaling
                    y = (
                        self.tiled_map.map_size.height * self.tiled_map.tile_size[1]
                        - cur_object.coordinates.y
                    ) * scaling

                    shape = (x + offset[0], y + offset[1])
                else:
                    sx = cur_object.coordinates.x * scaling + offset[0]
                    sy = (
                        self.tiled_map.map_size.height * self.tiled_map.tile_size[1]
                        - cur_object.coordinates.y
                    ) * scaling + offset[1]

                    ex = sx + cur_object.size.width * scaling
                    ey = sy - cur_object.size.height * scaling

                    p1 = (sx, sy)
                    p2 = (ex, sy)
                    p3 = (ex, ey)
                    p4 = (sx, ey)

                    shape = [p1, p2, p3, p4]
            elif isinstance(cur_object, pytiled_parser.tiled_object.Polygon) or isinstance(
                cur_object, pytiled_parser.tiled_object.Polyline
            ):
                points: list[Point2] = []
                shape = points
                for point in cur_object.points:
                    x = point.x + cur_object.coordinates.x
                    y = (self.height * self.tile_height) - (point.y + cur_object.coordinates.y)
                    point = (x + offset[0], y + offset[1])  # type: ignore
                    points.append(point)

                # If shape is a polyline, and it is closed, we need to
                # remove the duplicate end point
                if points[0][0] == points[-1][0] and points[0][1] == points[-1][1]:
                    points.pop()
            elif isinstance(cur_object, pytiled_parser.tiled_object.Ellipse):
                hw = cur_object.size.width / 2
                hh = cur_object.size.height / 2
                cx = cur_object.coordinates.x + hw
                cy = cur_object.coordinates.y + hh

                total_steps = 8
                angles = [step / total_steps * 2 * math.pi for step in range(total_steps)]
                points = []
                shape = points
                for angle in angles:
                    x = hw * math.cos(angle) + cx
                    y = -(hh * math.sin(angle) + cy)
                    point = (x + offset[0], y + offset[1])  # type: ignore
                    points.append(point)
            elif isinstance(cur_object, pytiled_parser.tiled_object.Text):
                pass
            else:
                continue

            if shape:
                tiled_object = TiledObject(
                    shape, cur_object.properties, cur_object.name, cur_object.class_
                )

                if not objects_list:
                    objects_list = []

                objects_list.append(tiled_object)

        return sprite_list or None, objects_list or None


def load_tilemap(
    map_file: str | Path,
    scaling: float = 1.0,
    layer_options: dict[str, dict[str, Any]] | None = None,
    use_spatial_hash: bool = False,
    hit_box_algorithm: HitBoxAlgorithm | None = None,
    offset: Vec2 = Vec2(0, 0),
    texture_atlas: DefaultTextureAtlas | None = None,
    lazy: bool = False,
) -> TileMap:
    """
    Given a .json map file, loads in and returns a `TileMap` object.

    A TileMap can be created directly using the classes `__init__` function.
    This function exists for ease of use.

    For more clarification on the layer_options key, see the `__init__` function
    of the `TileMap` class

    Args:
        map_file:
            The JSON map file.
        scaling:
            The global scaling to apply to all Sprite's within the map.
        use_spatial_hash:
            If set to True, this will make moving a sprite
            in the SpriteList slower, but it will speed up collision detection
            with items in the SpriteList. Great for doing collision detection
            with static walls/platforms.
        hit_box_algorithm:
            The hit box algorithm to use for collision detection.
        layer_options:
            Layer specific options for the map.
        offset:
            Can be used to offset the position of all sprites and objects
            within the map. This will be applied in addition to any offsets from Tiled. This value
            can be overridden with the layer_options dict.
        texture_atlas:
            A default texture atlas to use for the SpriteLists created by this map.
            If not supplied the global default atlas will be used.
        lazy:
            SpriteLists will be created lazily.
    """
    return TileMap(
        map_file=map_file,
        scaling=scaling,
        layer_options=layer_options,
        use_spatial_hash=use_spatial_hash,
        hit_box_algorithm=hit_box_algorithm,
        offset=offset,
        texture_atlas=texture_atlas,
        lazy=lazy,
    )



"""
Render some basic geometry with shaders.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.basic_renderer
"""

import math
import arcade
from arcade.gl import geometry

# Do the math to figure out our screen dimensions
WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Basic Renderer"


class GameView(arcade.Window):

    def __init__(self, width, height, title):
        """
        Set up the application.
        """
        super().__init__(width, height, title)
        self.color_program = self.ctx.program(
            vertex_shader="""
            #version 330

            in vec2 in_vert;

            void main() {
                gl_Position = vec4(in_vert, 0.0, 1.0);
            }
            """,
            fragment_shader="""
            #version 330

            uniform vec4 color;
            out vec4 out_color;

            void main() {
                out_color = color;
            }
            """,
        )
        self.uv_program = self.ctx.program(
            vertex_shader="""
            #version 330

            in vec2 in_vert;
            in vec2 in_uv;
            out vec2 v_uv;

            void main() {
                gl_Position = vec4(in_vert, 0.0, 1.0);
                v_uv = in_uv;
            }
            """,
            fragment_shader="""
            #version 330

            in vec2 v_uv;
            out vec4 out_color;

            void main() {
                out_color = vec4(v_uv, 0.0, 1.0);
            }
            """,
        )
        self.quad_1 = geometry.quad_2d(pos=(-0.5, 0.5))
        self.quad_2 = geometry.quad_2d(pos=(0.5, 0.5))

    def on_draw(self):
        self.clear()
        self.color_program["color"] = (
            (math.sin(self.time) + 1.0) / 2,
            (math.sin(self.time + 2) + 1.0) / 2,
            (math.sin(self.time + 3) + 1.0) / 2,
            1.0,
        )
        self.quad_1.render(self.color_program)
        self.quad_2.render(self.uv_program)


if __name__ == "__main__":
    GameView(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)
    arcade.run()



from array import array
from collections.abc import Generator
from contextlib import contextmanager

from PIL import Image
from pyglet.math import Vec2, Vec4
from typing_extensions import Self

import arcade
from arcade import Texture
from arcade.camera import CameraData, OrthographicProjectionData, OrthographicProjector
from arcade.color import TRANSPARENT_BLACK
from arcade.gl import BufferDescription, Framebuffer
from arcade.gui.nine_patch import NinePatchTexture
from arcade.types import LBWH, RGBA255, Point, Rect


class Surface:
    """Internal abstraction for widget rendering.

    Holds a :class:`arcade.gl.Framebuffer` and provides helper methods
    and properties for drawing to it.

    Args:
        size: The size of the surface in window coordinates
        position: The position of the surface in window
        pixel_ratio: The pixel scale of the window
    """

    def __init__(
        self,
        *,
        size: tuple[int, int],
        position: tuple[int, int] = (0, 0),
        pixel_ratio: float = 1.0,
    ):
        self.ctx = arcade.get_window().ctx
        self._size = size
        self._pos = position
        self._pixel_ratio = pixel_ratio
        self._pixelated = False
        self._area: Rect | None = None  # Cached area for the last draw call

        self.texture = self.ctx.texture(self.size_scaled, components=4)
        self.fbo: Framebuffer = self.ctx.framebuffer(color_attachments=[self.texture])
        self.fbo.clear()

        #: Blend modes for when we're drawing into the surface
        self.blend_func_render_into = (
            *self.ctx.BLEND_DEFAULT,
            *self.ctx.BLEND_ADDITIVE,
        )
        #: Blend mode for when we're drawing the surface
        self.blend_func_render = (
            *self.ctx.BLEND_DEFAULT,
            *self.ctx.BLEND_DEFAULT,
        )

        # 5 floats per vertex (pos 3f, tex 2f) with 4 vertices
        self._buffer = self.ctx.buffer(reserve=4 * 5 * 4)
        self._geometry = self.ctx.geometry(
            content=[BufferDescription(self._buffer, "3f 2f", ["in_pos", "in_uv"])],
            mode=self.ctx.TRIANGLE_STRIP,
        )
        self._program = self.ctx.load_program(
            vertex_shader=":system:shaders/gui/surface_vs.glsl",
            fragment_shader=":system:shaders/gui/surface_fs.glsl",
        )
        self._update_geometry()

        self._cam = OrthographicProjector(
            view=CameraData(),
            projection=OrthographicProjectionData(0.0, self.width, 0.0, self.height, -100, 100),
            viewport=LBWH(0, 0, self.width, self.height),
        )

    @property
    def position(self) -> Point:
        """Get or set the surface position"""
        return self._pos

    @position.setter
    def position(self, value):
        self._pos = value

    @property
    def size(self):
        """Size of the surface in window coordinates"""
        return self._size

    @property
    def size_scaled(self):
        """The physical size of the buffer"""
        return (
            int(self._size[0] * self._pixel_ratio),
            int(self._size[1] * self._pixel_ratio),
        )

    @property
    def pixel_ratio(self) -> float:
        """The pixel ratio of the surface"""
        return self._pixel_ratio

    @property
    def width(self) -> int:
        """Width of the surface"""
        return self._size[0]

    @property
    def height(self) -> int:
        """Height of the surface"""
        return self._size[1]

    def clear(self, color: RGBA255 = TRANSPARENT_BLACK):
        """Clear the surface"""
        self.fbo.clear(color=color)

    def draw_texture(
        self,
        x: float,
        y: float,
        width: float,
        height: float,
        tex: Texture | NinePatchTexture,
        angle: float = 0.0,
        alpha: int = 255,
    ):
        """Draw a texture to the surface.

        Args:
            x: The x coordinate of the texture.
            y: The y coordinate of the texture.
            width: The width of the texture.
            height: The height of the texture.
            tex: The texture to draw, also supports NinePatchTexture.
            angle: The angle of the texture.
            alpha: The alpha value of the texture.
        """
        if isinstance(tex, NinePatchTexture):
            if angle != 0.0:
                raise NotImplementedError(
                    f"Ninepatch does not support an angle != 0 yet, but got {angle}"
                )

            if alpha != 255:
                raise NotImplementedError(
                    f"Ninepatch does not support an alpha != 255 yet, but got {alpha}"
                )

            tex.draw_rect(rect=LBWH(0, 0, width, height))
        else:
            arcade.draw_texture_rect(
                tex, LBWH(x, y, width, height), angle=angle, alpha=alpha, pixelated=self._pixelated
            )

    def draw_sprite(self, x: float, y: float, width: float, height: float, sprite: arcade.Sprite):
        """Draw a sprite to the surface

        Args:
            x: The x coordinate of the sprite.
            y: The y coordinate of the sprite.
            width: The width of the sprite.
            height: The height of the sprite.
            sprite: The sprite to draw.
        """
        sprite.position = x + width // 2, y + height // 2
        sprite.width = width
        sprite.height = height
        arcade.draw_sprite(sprite, pixelated=self._pixelated)

    @contextmanager
    def activate(self) -> Generator[Self, None, None]:
        """Context manager for rendering safely to this :py:class:`Surface`.

        It does the following:

        #. Apply this surface's viewport, projection, and blend settings
        #. Allow any rendering to take place
        #. Restore the old OpenGL context settings

        Use it in ``with`` blocks like other managers:

        .. code-block:: python

           with surface.activate():
               # draw stuff here

        """
        # Set viewport and projection
        self.limit(LBWH(0, 0, *self.size))
        # Set blend function
        prev_blend_func = self.ctx.blend_func

        try:
            self.ctx.blend_func = self.blend_func_render_into
            with self.fbo.activate():
                yield self
        finally:
            # Restore blend function.
            self.ctx.blend_func = prev_blend_func

    def limit(self, rect: Rect | None = None):
        """Reduces the draw area to the given rect, or resets it to the full surface."""

        if rect is None:
            rect = LBWH(0, 0, *self.size)

        l, b, w, h = rect.lbwh
        w = max(w, 1)
        h = max(h, 1)

        # round to nearest pixel, to avoid off by 1-pixel errors in ui
        viewport_rect = LBWH(
            round(l * self._pixel_ratio),
            round(b * self._pixel_ratio),
            round(w * self._pixel_ratio),
            round(h * self._pixel_ratio),
        )
        self.fbo.viewport = viewport_rect.lbwh_int

        self._cam.projection.rect = LBWH(0, 0, w, h)
        self._cam.viewport = viewport_rect

        self._cam.use()

    def draw(
        self,
        area: Rect | None = None,
    ) -> None:
        """Draws the contents of the surface.

        The surface will be rendered at the configured ``position``
        and limited by the given ``area``. The area can be out of bounds.

        Args:
            area: Limit the area in the surface we're drawing
                (l, b, w, h)
        """
        self._update_geometry(area=area)

        # Set blend function
        blend_func = self.ctx.blend_func
        self.ctx.blend_func = self.blend_func_render

        # Handle the pixelated shortcut if filter is not set
        if self._pixelated:
            self.texture.filter = self.ctx.NEAREST, self.ctx.NEAREST
        else:
            self.texture.filter = self.ctx.LINEAR, self.ctx.LINEAR

        self.texture.use(0)
        self._geometry.render(self._program)

        # Restore blend function
        self.ctx.blend_func = blend_func

    def resize(self, *, size: tuple[int, int], pixel_ratio: float) -> None:
        """Resize the internal texture by re-allocating a new one

        Args:
            size: The new size in pixels (xy)
            pixel_ratio: The pixel scale of the window
        """
        # Texture re-allocation is expensive so we should block unnecessary calls.
        if self._size == size and self._pixel_ratio == pixel_ratio:
            return
        self._size = size
        self._pixel_ratio = pixel_ratio
        # Create new texture and fbo
        self.texture = self.ctx.texture(self.size_scaled, components=4)
        self.fbo = self.ctx.framebuffer(color_attachments=[self.texture])
        self.fbo.clear()

    def to_image(self) -> Image.Image:
        """Convert the surface to an PIL image"""
        return self.ctx.get_framebuffer_image(self.fbo)

    def _update_geometry(self, area: Rect | None = None) -> None:
        """
        Update the internal geometry of the surface mesh.

        The geometry is a triangle strip with 4 vertices.
        """
        if area is None:
            area = LBWH(0, 0, *self.size)

        if self._area == area:
            return
        self._area = area

        # Clamp the area inside the surface
        # This is the local area inside the surface
        _size = Vec2(*self.size)
        _pos = Vec2(*self.position)
        _area_pos = Vec2(area.left, area.bottom)
        _area_size = Vec2(area.width, area.height)

        b1 = _area_pos.clamp(Vec2(0.0), _size)
        end_point = _area_pos + _area_size
        b2 = end_point.clamp(Vec2(0.0), _size)
        b = b2 - b1
        l_area = Vec4(b1.x, b1.y, b.x, b.y)

        # Create the 4 corners of the rectangle
        # These are the final/global coordinates rendered
        p_ll = _pos + l_area.xy  # type: ignore
        p_lr = _pos + l_area.xy + Vec2(l_area.z, 0.0)  # type: ignore
        p_ul = _pos + l_area.xy + Vec2(0.0, l_area.w)  # type: ignore
        p_ur = _pos + l_area.xy + l_area.zw  # type: ignore

        # Calculate the UV coordinates
        bottom = l_area.y / _size.y
        left = l_area.x / _size.x
        top = (l_area.y + l_area.w) / _size.y
        right = (l_area.x + l_area.z) / _size.x

        # fmt: off
        vertices = array("f", (
            p_ll.x, p_ll.y, 0.0, left, bottom,
            p_lr.x, p_lr.y, 0.0, right, bottom,
            p_ul.x, p_ul.y, 0.0, left, top,
            p_ur.x, p_ur.y, 0.0, right, top,
        ))
        # fmt: on
        self._buffer.write(vertices)



from __future__ import annotations

from collections.abc import Generator
from contextlib import contextmanager
from math import atan2, cos, degrees, radians, sin
from typing import TYPE_CHECKING

from pyglet.math import Vec2, Vec3
from typing_extensions import Self

from arcade.camera.data_types import (
    DEFAULT_FAR,
    DEFAULT_NEAR_ORTHO,
    CameraData,
    OrthographicProjectionData,
    ZeroProjectionDimension,
)
from arcade.camera.projection_functions import (
    generate_orthographic_matrix,
    generate_view_matrix,
    project_orthographic,
    unproject_orthographic,
)
from arcade.types import LBWH, LRBT, XYWH, Point, Rect
from arcade.types.vector_like import Point2
from arcade.window_commands import get_window

if TYPE_CHECKING:
    from arcade.application import Window
    from arcade.gl import Framebuffer

__all__ = ["Camera2D"]


class Camera2D:
    """
    A simple orthographic camera.

    It provides properties to access every important variable for controlling the camera.
    3D properties such as pos, and up are constrained to a 2D plane. There is no access to the
    forward vector (as a property).

    There are also ease of use methods for matching the viewport and projector to the window size.

    Provides many helpful values:
        * The position and rotation or the camera
        * 8 positions along the edge of the camera's viewable area
        * the bounding box of the area the camera sees
        * Viewport, and Scissor box for controlling where to draw to

    .. warning:: Do not replace the ``camera_data`` and ``projection_data``
                 instances after initialization!

    Replacing the camera data and projection data may break controllers. Their
    contents are exposed via properties rather than directly to prevent this.

    Args:
        viewport:
            A ``Rect`` which defines the pixel bounds which the camera fits its image to.
            If the viewport is not 1:1 with the projection then positions in world space
            won't match pixels on screen.
        position:
            The 2D position of the camera in the XY plane.
        up:
            A 2D vector which describes which direction is up
            (defines the +Y-axis of the camera space).
        zoom:
            A scalar value which is inversely proportional to the size of the
            camera projection. i.e. a zoom of 2.0 halves the size of the projection,
            doubling the perceived size of objects.
        projection:
            A ``Rect`` which defines the world space
            bounds which the camera projects to the viewport.
        near:
            The near clipping plane of the camera.
        far:
            The far clipping plane of the camera.
        scissor:
            A ``Rect`` which will crop the camera's output to this area on screen.
            Unlike the viewport this has no influence on the visuals rendered with
            the camera only the area shown.
        render_target:
            The FrameBuffer that the camera may use. Warning if the target isn't the screen
            it won't automatically show up on screen.
        window:
            The Arcade Window to bind the camera to. Defaults to the currently active window.
    """

    def __init__(
        self,
        viewport: Rect | None = None,
        position: Point2 | None = None,
        up: tuple[float, float] = (0.0, 1.0),
        zoom: float = 1.0,
        projection: Rect | None = None,
        near: float = DEFAULT_NEAR_ORTHO,
        far: float = DEFAULT_FAR,
        *,
        scissor: Rect | None = None,
        render_target: Framebuffer | None = None,
        window: Window | None = None,
    ):
        self._window: Window = window or get_window()
        self.render_target: Framebuffer | None = render_target
        """
        An optional framebuffer to activate at the same time as
        the projection data, could be the screen, or an offscreen texture
        """

        # We don't want to force people to use a render target,
        # but we need to have some form of default size.
        render_target = render_target or self._window.ctx.screen
        viewport = viewport or LBWH(*render_target.viewport)
        width, height = viewport.size
        half_width = width / 2
        half_height = height / 2

        # Unpack projection, but only validate when it's given directly
        left, right, bottom, top = (
            (-half_width, half_width, -half_height, half_height)
            if projection is None
            else projection.lrbt
        )

        if projection is not None:
            if left == right:
                raise ZeroProjectionDimension(
                    f"projection width is 0 due to equal {left=} and {right=} values"
                )
            if bottom == top:
                raise ZeroProjectionDimension(
                    f"projection height is 0 due to equal {bottom=} and {top=}"
                )
        if near == far:
            raise ZeroProjectionDimension(
                f"projection depth is 0 due to equal {near=} and {far=} values"
            )

        pos_x = position[0] if position is not None else half_width
        pos_y = position[1] if position is not None else half_height
        self._camera_data = CameraData(
            position=(pos_x, pos_y, 0.0),
            up=(up[0], up[1], 0.0),
            forward=(0.0, 0.0, -1.0),
            zoom=zoom,
        )
        self._projection_data: OrthographicProjectionData = OrthographicProjectionData(
            left=left, right=right, top=top, bottom=bottom, near=near, far=far
        )

        self.viewport: Rect = viewport or LRBT(0, 0, width, height)
        """
        A rect which describes how the final projection should be mapped
        from unit-space. defaults to the size of the render_target or window
        """

        self.scissor: Rect | None = scissor
        """
        An optional rect which describes what pixels of the active render
        target should be drawn to when undefined the viewport rect is used.
        """

    @classmethod
    def from_camera_data(
        cls,
        *,
        camera_data: CameraData | None = None,
        projection_data: OrthographicProjectionData | None = None,
        render_target: Framebuffer | None = None,
        viewport: Rect | None = None,
        scissor: Rect | None = None,
        window: Window | None = None,
    ) -> Self:
        """
        Make a ``Camera2D`` directly from data objects.

        This :py:class:`classmethod` allows advanced users to:

        #. skip or replace the default validation
        #. share ``camera_data`` or ``projection_data`` between cameras

        .. warning:: Be careful when sharing data objects!
                    **Any** action on a camera which changes a shared
                    object changes it for **every** camera which uses
                    the same object.

        .. list-table::
          :header-rows: 1

          * - Shared Value
            - Example Use(s)
          * - ``camera_data``
            - Mini-maps, reflection, and ghosting effects.
          * - ``projection_data``
            - Simplified rendering configuration
          * - ``render_target``
            - Complex rendering setups

        Args:
            camera_data:
                A :py:class:`~arcade.camera.CameraData`
                describing the position, up, forward and zoom.
            projection_data:
                A :py:class:`~arcade.camera.OrthographicProjectionData`
                which describes the left, right, top, bottom, far, near
                planes and the viewport for an orthographic projection.
            render_target:
                A non-screen :py:class:`~arcade.gl.framebuffer.Framebuffer` for this
                camera to draw into. When specified,

                * nothing will draw directly to the screen
                * the buffer's internal viewport will be ignored

            viewport:
                A viewport as a :py:class:`~arcade.types.rect.Rect`.
                This overrides any viewport the ``render_target`` may have.
            scissor:
                The OpenGL scissor box to use when drawing.
            window: The Arcade Window to bind the camera to.
                Defaults to the currently active window.
        """

        if projection_data:
            left, right = projection_data.left, projection_data.right
            if projection_data.left == projection_data.right:
                raise ZeroProjectionDimension(
                    f"projection width is 0 due to equal {left=}and {right=} values"
                )
            bottom, top = projection_data.bottom, projection_data.top
            if bottom == top:
                raise ZeroProjectionDimension(
                    f"projection height is 0 due to equal {bottom=}and {top=}"
                )
            near, far = projection_data.near, projection_data.far
            if near == far:
                raise ZeroProjectionDimension(
                    f"projection depth is 0 due to equal {near=}and {far=} values"
                )

        # build a new camera with defaults and then apply the provided camera objects.
        new_camera = cls(
            render_target=render_target, window=window, viewport=viewport, scissor=scissor
        )

        if camera_data is not None:
            new_camera._camera_data = camera_data
        if projection_data is not None:
            new_camera._projection_data = projection_data

        return new_camera

    def use(self) -> None:
        """
        Set internal projector as window projector,
        and set the projection and view matrix.
        call every time you want to 'look through' this camera.

        If you want to use a 'with' block use activate() instead.
        """
        if self.render_target is not None:
            self.render_target.use()
        self._window.current_camera = self

        _projection = generate_orthographic_matrix(self.projection_data, self.zoom)
        _view = generate_view_matrix(self.view_data)

        self._window.ctx.viewport = self.viewport.lbwh_int
        self._window.ctx.scissor = None if not self.scissor else self.scissor.lbwh_int
        self._window.projection = _projection
        self._window.view = _view

    @contextmanager
    def activate(self) -> Generator[Self, None, None]:
        """
        Set internal projector as window projector,
        and set the projection and view matrix.

        This method works with 'with' blocks.
        After using this method it automatically resets
        the projector to the one previously in use.
        """
        previous_projection = self._window.current_camera
        previous_framebuffer = self._window.ctx.active_framebuffer
        try:
            self.use()
            yield self
        finally:
            previous_framebuffer.use()
            previous_projection.use()

    def project(self, world_coordinate: Point) -> Vec2:
        """
        Take a Vec2 or Vec3 of coordinates and return the related screen coordinate
        """
        _projection = generate_orthographic_matrix(self.projection_data, self.zoom)
        _view = generate_view_matrix(self.view_data)

        return project_orthographic(
            world_coordinate,
            self.viewport.lbwh_int,
            _view,
            _projection,
        )

    def unproject(self, screen_coordinate: Point) -> Vec3:
        """
        Take in a pixel coordinate from within
        the range of the window size and returns
        the world space coordinates.

        Essentially reverses the effects of the projector.

        Args:
            screen_coordinate: A 2D or 3D position in pixels from the bottom left of the screen.
        Returns:
            A 3D vector in world space (same as sprites).
            perfect for finding if the mouse overlaps with a sprite or ui element irrespective
            of the camera.
        """

        _projection = generate_orthographic_matrix(self.projection_data, self.zoom)
        _view = generate_view_matrix(self.view_data)
        return unproject_orthographic(screen_coordinate, self.viewport.lbwh_int, _view, _projection)

    def equalise(self) -> None:
        """
        Forces the projection to match the size of the viewport.
        When matching the projection to the viewport the method keeps
        the projections center in the same relative place.
        """
        x, y = self._projection_data.rect.x, self._projection_data.rect.y
        self._projection_data.rect = XYWH(x, y, self.viewport_width, self.viewport_height)

    def match_window(
        self,
        viewport: bool = True,
        projection: bool = True,
        scissor: bool = True,
        position: bool = False,
        aspect: float | None = None,
    ) -> None:
        """
        Sets the viewport to the size of the window.
        Should be called when the window is resized.

        Args:
            viewport: Flag whether to equalise the viewport to the value.
            projection: Flag whether to also equalize the projection to the viewport.
                On by default
            scissor: Flag whether to also equalize the scissor box to the viewport.
                On by default
            position: Flag whether to also center the camera to the viewport.
                Off by default
            aspect: The ratio between width and height that the viewport should
                be constrained to. If unset then the viewport just matches the window
                size. The aspect ratio describes how much larger the width should be
                compared to the height. i.e. for an aspect ratio of ``4:3`` you should
                input ``4.0/3.0`` or ``1.33333...``. Cannot be equal to zero.
        """
        self.update_values(
            self._window.rect,
            viewport=viewport,
            projection=projection,
            scissor=scissor,
            position=position,
            aspect=aspect,
        )

    def match_target(
        self,
        viewport: bool = True,
        projection: bool = True,
        scissor: bool = True,
        position: bool = False,
        aspect: float | None = None,
    ) -> None:
        """
        Sets the viewport to the size of the Camera2D's render target.

        Args:
            viewport: Flag whether to equalize the viewport to the area of the render target
            projection: Flag whether to equalize the size of the projection to
                match the render target.
                The projection center stays fixed, and the new projection matches only in size.
            scissor: Flag whether to update the scissor value.
            position: Flag whether to also center the camera to the value.
                Off by default
            aspect: The ratio between width and height that the value should
                be constrained to. i.e. for an aspect ratio of ``4:3`` you should
                input ``4.0/3.0`` or ``1.33333...``. Cannot be equal to zero.
                If unset then the value will not be updated.
        Raises:
            ValueError: Will be raised if the Camera2D was has no render target.
        """
        if self.render_target is None:
            raise ValueError(
                "Tried to match a non-exsistant render target. Please use `match_window` instead"
            )

        self.update_values(
            LRBT(*self.render_target.viewport),
            viewport,
            projection,
            scissor,
            position,
            aspect=aspect,
        )

    def update_values(
        self,
        value: Rect,
        viewport: bool = True,
        projection: bool = True,
        scissor: bool = True,
        position: bool = False,
        aspect: float | None = None,
    ):
        """
        Convenience method for updating the viewport, projection, position
        and a few others with the same value.

        Args:
            value: The rect that the values will be derived from.
            viewport: Flag whether to equalize the viewport to the value.
            projection: Flag whether to equalize the size of the projection to match the value.
                The projection center stays fixed, and the new projection matches only in size.
            scissor: Flag whether to update the scissor value.
            position: Flag whether to also center the camera to the value.
                Off by default
            aspect: The ratio between width and height that the value should
                be constrained to. i.e. for an aspect ratio of ``4:3`` you should
                input ``4.0/3.0`` or ``1.33333...``. Cannot be equal to zero.
                If unset then the value will not be updated.
        """
        if aspect is not None:
            if value.height * aspect < value.width:
                w = value.height * aspect
                h = value.height
            else:
                w = value.width
                h = value.width / aspect
            value = XYWH(value.x, value.y, w, h)

        if viewport:
            self.viewport = value

        if projection:
            x, y = self._projection_data.rect.x, self._projection_data.rect.y
            self._projection_data.rect = XYWH(x, y, value.width, value.height)

        if scissor and self.scissor:
            self.scissor = value

        if position:
            self.position = value.center

    def aabb(self) -> Rect:
        """
        Retrieve the axis-aligned bounds box of the camera's view area.

        If the camera isn't rotated , this will be precisely the view area,
        but it will cover a larger area when it is rotated. Useful for CPU culling
        """
        up = self._camera_data.up
        ux, uy, *_ = up
        rx, ry = uy, -ux  # up x Z'

        l, r, b, t = self.viewport.lrbt
        x, y = self.position

        x_points = (
            x + ux * t + rx * l,  # top left
            x + ux * t + rx * r,  # top right
            x + ux * b + rx * l,  # bottom left
            x + ux * b + rx * r,  # bottom right
        )
        y_points = (
            y + uy * t + ry * l,  # top left
            y + uy * t + ry * r,  # top right
            y + uy * b + ry * l,  # bottom left
            y + uy * b + ry * r,  # bottom right
        )

        left = min(x_points)
        right = max(x_points)
        bottom = min(y_points)
        top = max(y_points)
        return LRBT(left=left, right=right, bottom=bottom, top=top)

    def point_in_view(self, point: Point2) -> bool:
        # TODO test
        """
        Take a 2D point in the world, and return whether the point is inside the
        visible area of the camera.
        """
        # This is unwrapped from standard Vec2 operations,
        # The construction and garbage collection of the vectors would
        # increase this method's cost by ~4x

        pos = self.position
        diff = point[0] - pos[0], point[1] - pos[1]

        up = self._camera_data.up

        h_width = self.width / 2.0
        h_height = self.height / 2.0

        dot_x = up[1] * diff[0] - up[0] * diff[1]
        dot_y = up[0] * diff[0] + up[1] * diff[1]

        return abs(dot_x) <= h_width and abs(dot_y) <= h_height

    @property
    def view_data(self) -> CameraData:
        """The view data for the camera.

        This includes:

        * the position
        * forward vector
        * up direction
        * zoom.

        Camera controllers use this property.
        """
        return self._camera_data

    @property
    def projection_data(self) -> OrthographicProjectionData:
        """The projection data for the camera.

        This is an Orthographic projection. with a
        right, left, top, bottom, near, and far value.

        An easy way to understand the use of the projection is
        that the right value of the projection tells the
        camera what value will be at the right most
        pixel in the viewport.

        Due to the view data having a zoom component
        most use cases will only change the projection
        on screen resize.
        """
        return self._projection_data

    @property
    def position(self) -> Vec2:
        """The 2D world position of the camera along the X and Y axes."""
        return Vec2(self._camera_data.position[0], self._camera_data.position[1])

    # Setter with different signature will cause mypy issues
    # https://github.com/python/mypy/issues/3004
    @position.setter
    def position(self, _pos: Point) -> None:
        x, y, *_z = _pos
        z = self._camera_data.position[2] if not _z else _z[0]
        self._camera_data.position = (x, y, z)

    @property
    def projection(self) -> Rect:
        """Get/set the left, right, bottom, and top projection values.

        These are world space values which control how the camera
        projects the world onto the pixel space of the current
        viewport area.

        .. note:: this IS scaled by zoom.
                  If this isn't what you want,
                  you have to calculate the value manually from projection_data

        .. warning:: The axis values cannot be equal!

                     * ``left`` cannot equal ``right``
                     * ``bottom`` cannot equal ``top``

        This property raises a :py:class:`~arcade.camera.data_types.ZeroProjectionDimension`
        exception if any axis pairs are equal. You can handle this
        exception as a :py:class:`ValueError`.
        """

        return self._projection_data.rect / self._camera_data.zoom

    @projection.setter
    def projection(self, value: Rect) -> None:
        # Unpack and validate
        if not value:
            raise ZeroProjectionDimension(f"Projection area is 0, {value.lrbt}")

        _z = self._camera_data.zoom

        # Modify the projection data itself.
        self._projection_data.rect = value * _z

    @property
    def width(self) -> float:
        """
        The width of the projection from left to right.
        This is in world space coordinates not pixel coordinates.

        .. note:: this IS scaled by zoom.
                  If this isn't what you want,
                  you have to calculate the value manually from projection_data
        """
        return (self._projection_data.right - self._projection_data.left) / self._camera_data.zoom

    @width.setter
    def width(self, new_width: float) -> None:
        w = self.width
        l = self.left / w  # Normalised Projection left
        r = self.right / w  # Normalised Projection Right

        self.left = l * new_width
        self.right = r * new_width

    @property
    def height(self) -> float:
        """
        The height of the projection from bottom to top.
        This is in world space coordinates not pixel coordinates.

        .. note:: this IS scaled by zoom.
                  If this isn't what you want,
                  you have to calculate the value manually from projection_data
        """
        return (self._projection_data.top - self._projection_data.bottom) / self._camera_data.zoom

    @height.setter
    def height(self, new_height: float) -> None:
        h = self.height
        b = self.bottom / h  # Normalised Projection Bottom
        t = self.top / h  # Normalised Projection Top

        self.bottom = b * new_height
        self.top = t * new_height

    @property
    def left(self) -> float:
        """
        The left edge of the projection in world space.
        This is not adjusted with the camera position.

        .. note:: this IS scaled by zoom.
                  If this isn't what you want,
                  you have to calculate the value manually from projection_data
        """
        return self._projection_data.left / self._camera_data.zoom

    @left.setter
    def left(self, new_left: float) -> None:
        self._projection_data.left = new_left * self._camera_data.zoom

    @property
    def right(self) -> float:
        """
        The right edge of the projection in world space.
        This is not adjusted with the camera position.

        .. note:: this IS scaled by zoom.
                  If this isn't what you want,
                  you have to calculate the value manually from projection_data
        """
        return self._projection_data.right / self._camera_data.zoom

    @right.setter
    def right(self, new_right: float) -> None:
        self._projection_data.right = new_right * self._camera_data.zoom

    @property
    def bottom(self) -> float:
        """
        The bottom edge of the projection in world space.
        This is not adjusted with the camera position.

        .. note:: this IS scaled by zoom.
                  If this isn't what you want,
                  you have to calculate the value manually from projection_data
        """
        return self._projection_data.bottom / self._camera_data.zoom

    @bottom.setter
    def bottom(self, new_bottom: float) -> None:
        self._projection_data.bottom = new_bottom * self._camera_data.zoom

    @property
    def top(self) -> float:
        """
        The top edge of the projection in world space.
        This is not adjusted with the camera position.

        .. note:: this IS scaled by zoom.
                  If this isn't what you want,
                  you have to calculate the value manually from projection_data
        """
        return self._projection_data.top / self._camera_data.zoom

    @top.setter
    def top(self, new_top: float) -> None:
        self._projection_data.top = new_top * self._camera_data.zoom

    @property
    def projection_near(self) -> float:
        """
        The near plane of the projection in world space.
        This is not adjusted with the camera position.

        .. note:: this IS NOT scaled by zoom.
        """
        return self._projection_data.near

    @projection_near.setter
    def projection_near(self, new_near: float) -> None:
        self._projection_data.near = new_near

    @property
    def projection_far(self) -> float:
        """
        The far plane of the projection in world space.
        This is not adjusted with the camera position.

        .. note:: this IS NOT scaled by zoom.
        """
        return self._projection_data.far

    @projection_far.setter
    def projection_far(self, new_far: float) -> None:
        self._projection_data.far = new_far

    @property
    def viewport_width(self) -> int:
        """
        The width of the viewport.
        Defines the number of pixels drawn too horizontally.
        """
        return int(self.viewport.width)

    @viewport_width.setter
    def viewport_width(self, new_width: int) -> None:
        self.viewport = self.viewport.resize(new_width, anchor=Vec2(0.0, 0.0))

    @property
    def viewport_height(self) -> int:
        """
        The height of the viewport.
        Defines the number of pixels drawn too vertically.
        """
        return int(self.viewport.height)

    @viewport_height.setter
    def viewport_height(self, new_height: int) -> None:
        self.viewport = self.viewport.resize(height=new_height, anchor=Vec2(0.0, 0.0))

    @property
    def viewport_left(self) -> int:
        """
        The left most pixel drawn to on the X axis.
        """
        return int(self.viewport.left)

    @viewport_left.setter
    def viewport_left(self, new_left: int) -> None:
        """
        Set the left most pixel drawn to.
        This moves the position of the viewport, and does not change the size.
        """
        self.viewport = self.viewport.align_left(new_left)

    @property
    def viewport_right(self) -> int:
        """
        The right most pixel drawn to on the X axis.
        """
        return int(self.viewport.right)

    @viewport_right.setter
    def viewport_right(self, new_right: int) -> None:
        """
        Set the right most pixel drawn to.
        This moves the position of the viewport, and does not change the size.
        """
        self.viewport = self.viewport.align_right(new_right)

    @property
    def viewport_bottom(self) -> int:
        """
        The bottom most pixel drawn to on the Y axis.
        """
        return int(self.viewport.bottom)

    @viewport_bottom.setter
    def viewport_bottom(self, new_bottom: int) -> None:
        """
        Set the bottom most pixel drawn to.
        This moves the position of the viewport, and does not change the size.
        """
        self.viewport = self.viewport.align_bottom(new_bottom)

    @property
    def viewport_top(self) -> int:
        """
        The top most pixel drawn to on the Y axis.
        """
        return int(self.viewport.top)

    @viewport_top.setter
    def viewport_top(self, new_top: int) -> None:
        """
        Set the top most pixel drawn to.
        This moves the position of the viewport, and does not change the size.
        """
        self.viewport = self.viewport.align_top(new_top)

    @property
    def up(self) -> Vec2:
        """
        A 2D vector which describes what is mapped
        to the +Y direction on screen.
        This is equivalent to rotating the screen.
        The base vector is 3D, but this
        camera only provides a 2D view.
        """
        return Vec2(self._camera_data.up[0], self._camera_data.up[1])

    @up.setter
    def up(self, _up: Point2) -> None:
        """
        Set the 2D vector which describes what is
        mapped to the +Y direction on screen.
        This is equivalent to rotating the screen.
        The base vector is 3D, but this
        camera only provides a 2D view.

        .. warning:: This is assumed to be normalized (length 1.0)
        """
        x, y = _up
        self._camera_data.up = (x, y, 0.0)

    @property
    def angle(self) -> float:
        """
        An angle representation of the 2D UP vector.
        This starts with 0 degrees as [0, 1] rotating
        clock-wise.
        """
        # We rotate counter clockwise by 90 degrees because we want 0 deg to be directly up
        angle = degrees(atan2(self._camera_data.up[1], self._camera_data.up[0])) - 90.0
        if angle <= 0.0:
            angle += 360.0
        return 360 - angle

    @angle.setter
    def angle(self, value: float) -> None:
        """
        Set the 2D UP vector using an angle.
        This starts with 0 degrees as [0, 1]
        rotating clock-wise.
        """
        _r = radians(90.0 - value)
        # Note that this is flipped as we want 0 degrees to be vert.
        self._camera_data.up = (cos(_r), sin(_r), 0.0)

    @property
    def zoom(self) -> float:
        """
        A scalar value which describes
        how much the projection should
        be scaled towards from its center.

        A value of 2.0 causes the projection
        to be half its original size.
        This causes sprites to appear 2.0x larger.
        """
        return self._camera_data.zoom

    @zoom.setter
    def zoom(self, _zoom: float) -> None:
        """
        Set the scalar value which describes
        how much the projection should
        be scaled towards from its center.

        A value of 2.0 causes the projection
        to be half its original size.
        This causes sprites to appear 2.0x larger.
        """
        self._camera_data.zoom = _zoom

    # top_left
    @property
    def top_left(self) -> Vec2:
        """Get the top left most corner the camera can see"""
        pos = self.position
        ux, uy, *_ = self._camera_data.up
        rx, ry = uy, -ux

        top = self.top
        left = self.left

        return Vec2(pos.x + ux * top + rx * left, pos.y + uy * top + ry * left)

    @top_left.setter
    def top_left(self, new_corner: Point2):
        ux, uy, *_ = self._camera_data.up
        rx, ry = uy, -ux

        top = self.top
        left = self.left

        x, y = new_corner
        self.position = (x - ux * top - rx * left, y - uy * top - ry * left)  # type: ignore

    # top_center
    @property
    def top_center(self) -> Vec2:
        """Get the top most position the camera can see"""
        pos = self.position

        ux, uy, *_ = self._camera_data.up
        top = self.top
        return Vec2(pos.x + ux * top, pos.y + uy * top)

    @top_center.setter
    def top_center(self, new_top: Point2):
        ux, uy, *_ = self._camera_data.up
        top = self.top

        x, y = new_top
        self.position = x - ux * top, y - uy * top  # type: ignore

    # top_right
    @property
    def top_right(self) -> Vec2:
        """Get the top right most corner the camera can see"""
        pos = self.position
        ux, uy, *_ = self._camera_data.up
        rx, ry = uy, -ux

        top = self.top
        right = self.right

        return Vec2(pos.x + ux * top + rx * right, pos.y + uy * top + ry * right)

    @top_right.setter
    def top_right(self, new_corner: Point2):
        ux, uy, *_ = self._camera_data.up
        rx, ry = uy, -ux

        top = self.top
        right = self.right

        x, y = new_corner
        self.position = (x - ux * top - rx * right, y - uy * top - ry * right)  # type: ignore

    # center_right
    @property
    def center_right(self) -> Vec2:
        """Get the right most point the camera can see"""
        pos = self.position
        ux, uy, *_ = self._camera_data.up
        right = self.right
        return Vec2(pos.x + uy * right, pos.y - ux * right)

    @center_right.setter
    def center_right(self, new_right: Point2):
        ux, uy, *_ = self._camera_data.up
        right = self.right

        x, y = new_right
        self.position = x - uy * right, y + ux * right  # type: ignore

    # bottom_right
    @property
    def bottom_right(self) -> Vec2:
        """Get the bottom right most corner the camera can see"""
        pos = self.position
        ux, uy, *_ = self._camera_data.up
        rx, ry = uy, -ux

        bottom = self.bottom
        right = self.right
        return Vec2(pos.x + ux * bottom + rx * right, pos.y + uy * bottom + ry * right)

    @bottom_right.setter
    def bottom_right(self, new_corner: Point2):
        ux, uy, *_ = self._camera_data.up
        rx, ry = uy, -ux

        bottom = self.bottom
        right = self.right

        x, y = new_corner
        self.position = (
            x - ux * bottom - rx * right,
            y - uy * bottom - ry * right,
        )  # type: ignore

    # bottom_center
    @property
    def bottom_center(self) -> Vec2:
        """Get the bottom most position the camera can see"""
        pos = self.position
        ux, uy, *_ = self._camera_data.up
        bottom = self.bottom

        return Vec2(pos.x + ux * bottom, pos.y + uy * bottom)

    @bottom_center.setter
    def bottom_center(self, new_bottom: Point2):
        ux, uy, *_ = self._camera_data.up
        bottom = self.bottom

        x, y = new_bottom
        self.position = x - ux * bottom, y - uy * bottom  # type: ignore

    # bottom_left
    @property
    def bottom_left(self) -> Vec2:
        """Get the bottom left most corner the camera can see"""
        pos = self.position
        ux, uy, *_ = self._camera_data.up
        rx, ry = uy, -ux

        bottom = self.bottom
        left = self.left

        return Vec2(pos.x + ux * bottom + rx * left, pos.y + uy * bottom + ry * left)

    @bottom_left.setter
    def bottom_left(self, new_corner: Point2):
        ux, uy, *_ = self._camera_data.up
        rx, ry = uy, -ux

        bottom = self.bottom
        left = self.left

        x, y = new_corner
        self.position = x - ux * bottom - rx * left, y - uy * bottom - ry * left  # type: ignore

    # center_left
    @property
    def center_left(self) -> Vec2:
        """Get the left most point the camera can see"""
        pos = self.position
        ux, uy, *_ = self._camera_data.up
        left = self.left
        return Vec2(pos.x + uy * left, pos.y - ux * left)

    @center_left.setter
    def center_left(self, new_left: Point2):
        ux, uy, *_ = self._camera_data.up
        left = self.left

        x, y = new_left
        self.position = Vec2(x - uy * left, y + ux * left)



"""
Experimental stuff. API may change.
"""

from .shadertoy import Shadertoy, ShadertoyBuffer, ShadertoyBase
from .crt_filter import CRTFilter
from .bloom_filter import BloomFilter


__all__ = ["Shadertoy", "ShadertoyBuffer", "ShadertoyBase", "CRTFilter", "BloomFilter"]



"""
CHIP-8 is an interpreted minimalist programming language that was designed by
Joseph Weisbecker in the 1970s for use on the RCA COSMAC VIP computer.

This example renders a 64 x 32 pixel display. The pixel format is using
bits to represent what pixel is enabled/disabled so the byte size of
the buffer is 8 x 4 = 32 bytes.

References:
http://mattmik.com/files/chip8/mastering/chip8.html
https://github.com/JohnEarnest/Octo

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.chip8_display
"""

from array import array

import arcade
from arcade.gl import geometry

# Do the math to figure out our screen dimensions
WINDOW_WIDTH = 64 * 10
WINDOW_HEIGHT = 32 * 10
WINDOW_TITLE = "CHIP-8 Screen"


class GameView(arcade.Window):

    def __init__(self, width, height, title):
        """CHIP-8 Screen"""
        super().__init__(width, height, title)
        self.program = self.ctx.program(
            vertex_shader="""
            #version 330

            uniform mat4 projection;

            in vec2 in_vert;
            in vec2 in_uv;
            out vec2 v_uv;

            void main() {
                gl_Position = projection * vec4(in_vert, 0.0, 1.0);
                v_uv = in_uv;
            }
            """,
            fragment_shader="""
            #version 330

            // Unsigned integer sampler for reading uint data from texture
            uniform usampler2D screen;

            in vec2 v_uv;
            out vec4 out_color;

            void main() {
                // Calculate the bit position on the x axis
                uint bit_pos = uint(round((v_uv.x * 64) - 0.5)) % 8u;
                // Create bit mask we can AND the fragment with to extract the pixel value
                uint flag = uint(pow(2u, 7u - bit_pos));
                // Read the fragment value (We reverse the y axis here as well)
                uint frag = texture(screen, v_uv * vec2(1.0, -1.0)).r;
                // Write the pixel value. Values above 1 will be clamped to 1.
                out_color = vec4(vec3(frag & flag), 1.0);
            }
            """,
        )
        # 8 x 4
        self.program["projection"] = self.projection
        self.program["screen"] = 0
        b = 0  # border to test scale
        self.quad = geometry.screen_rectangle(b, b, WINDOW_WIDTH - b * 2, WINDOW_HEIGHT - b * 2)
        self.texture = self.ctx.texture((8, 32), components=1, dtype="i1")
        # fmt: off
        self.texture.write(array(
            'B',
            [
                # 64 x 32 screen
                0xAA, 0xAA, 0xF0, 0xF0, 0xFF, 0x00, 0xFF, 0xFF,
                0x55, 0x55, 0xF0, 0xF0, 0xFF, 0x00, 0xFF, 0xFF,
                0xAA, 0xAA, 0xF0, 0xF0, 0xFF, 0x00, 0xFF, 0xFF,
                0x55, 0x55, 0xF0, 0xF0, 0xFF, 0x00, 0xFF, 0xFF,
                0xAA, 0xAA, 0x0F, 0x0F, 0xFF, 0x00, 0xFF, 0xFF,
                0x55, 0x55, 0x0F, 0x0F, 0xFF, 0x00, 0xFF, 0xFF,
                0xAA, 0xAA, 0x0F, 0x0F, 0xFF, 0x00, 0xFF, 0xFF,
                0x55, 0x55, 0x0F, 0x0F, 0xFF, 0x00, 0xFF, 0xFF,

                0xAA, 0xAA, 0xF0, 0xF0, 0x00, 0xFF, 0xFF, 0xFF,
                0x55, 0x55, 0xF0, 0xF0, 0x00, 0xFF, 0xFF, 0xFF,
                0xAA, 0xAA, 0xF0, 0xF0, 0x00, 0xFF, 0xFF, 0xFF,
                0x55, 0x55, 0xF0, 0xF0, 0x00, 0xFF, 0xFF, 0xFF,
                0xAA, 0xAA, 0x0F, 0x0F, 0x00, 0xFF, 0xFF, 0xFF,
                0x55, 0x55, 0x0F, 0x0F, 0x00, 0xFF, 0xFF, 0xFF,
                0xAA, 0xAA, 0x0F, 0x0F, 0x00, 0xFF, 0xFF, 0xFF,
                0x55, 0x55, 0x0F, 0x0F, 0x00, 0xFF, 0xFF, 0xFF,

                0xAA, 0xAA, 0xCC, 0x99, 0xbd, 0x63, 0x80, 0x00,
                0x55, 0x55, 0xCC, 0x99, 0xbd, 0x63, 0x40, 0x00,
                0xAA, 0xAA, 0xCC, 0x99, 0xbd, 0x63, 0x20, 0x00,
                0x55, 0x55, 0xCC, 0x99, 0xbd, 0x63, 0x10, 0x00,
                0xAA, 0xAA, 0xCC, 0x99, 0xbd, 0x63, 0x08, 0x00,
                0x55, 0x55, 0xCC, 0x99, 0xbd, 0x63, 0x04, 0x00,
                0xAA, 0xAA, 0xCC, 0x99, 0xbd, 0x63, 0x02, 0x00,
                0x55, 0x55, 0xCC, 0x99, 0xbd, 0x63, 0x01, 0x00,

                0xAA, 0xAA, 0xCC, 0x99, 0xbd, 0x63, 0x00, 0x80,
                0x55, 0x55, 0xCC, 0x99, 0xbd, 0x63, 0x00, 0x40,
                0xAA, 0xAA, 0xCC, 0x99, 0xbd, 0x63, 0x00, 0x20,
                0x55, 0x55, 0xCC, 0x99, 0xbd, 0x63, 0x00, 0x10,
                0xAA, 0xAA, 0xCC, 0x99, 0xbd, 0x63, 0x00, 0x08,
                0x55, 0x55, 0xCC, 0x99, 0xbd, 0x63, 0x00, 0x04,
                0xAA, 0xAA, 0xCC, 0x99, 0xbd, 0x63, 0x00, 0x02,
                0x55, 0x55, 0xCC, 0x99, 0xbd, 0x63, 0x00, 0x01,
            ]
        ))
        # fmt: on

    def on_draw(self):
        self.clear()
        self.texture.use(0)
        self.quad.render(self.program)


if __name__ == "__main__":
    GameView(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)
    arcade.run()



"""
Rotating Sprites Around Points

Two minimal examples demonstrating how to rotate sprites around points
and how you might apply them in a game.

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_rotate_around_point
"""
import arcade
from arcade.math import rotate_point

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
QUARTER_WIDTH = WINDOW_WIDTH // 4
HALF_HEIGHT = WINDOW_HEIGHT // 2


WINDOW_TITLE = "Rotating Sprites Around Points"


class RotatingSprite(arcade.Sprite):
    """
    This sprite subclass implements a generic rotate_around_point method.
    """

    def rotate_around_point(self, point, degrees, change_angle=True):
        """
        Rotate the sprite around a point by the set amount of degrees

        You could remove the change_angle keyword and/or angle change
        if you know that sprites will always or never change angle.

        Args:
            point:
                The point that the sprite will rotate about
            degrees:
                How many degrees to rotate the sprite
            change_angle:
                Whether the sprite's angle should also be adjusted.
        """

        # If change_angle is true, change the sprite's angle
        if change_angle:
            self.angle += degrees

        # Move the sprite along a circle centered on the point by degrees
        self.position = rotate_point(
            self.center_x, self.center_y,
            point[0], point[1], degrees)


class GameView(arcade.View):

    def __init__(self):
        super().__init__()

        self.sprites = arcade.SpriteList()

        # This example is based off spinning lines of fire from Mario games
        # See https://www.mariowiki.com/Fire_Bar for more information
        self.left_rotating_laser_sprite = RotatingSprite(
            ":resources:images/space_shooter/laserBlue01.png",
            center_x=QUARTER_WIDTH + 26, center_y=HALF_HEIGHT)

        self.laser_base_sprite = arcade.Sprite(
            ":resources:images/tiles/boxCrate.png", scale=0.25,
            center_x=QUARTER_WIDTH, center_y=HALF_HEIGHT)

        self.laser_text = arcade.Text(
            "change_angle = True",
            QUARTER_WIDTH, WINDOW_HEIGHT // 2 - 150,
            anchor_x='center')

        # This example demonstrates how to make platforms rotate around a point
        self.right_rotating_platform_sprite = RotatingSprite(
            ":resources:images/tiles/grassHalf.png", scale=0.25,
            center_x=3 * QUARTER_WIDTH + 50, center_y=HALF_HEIGHT)

        self.platform_base_sprite = arcade.Sprite(
            ":resources:images/tiles/boxCrate.png", scale=0.25,
            center_x=3 * QUARTER_WIDTH, center_y=HALF_HEIGHT)

        self.platform_text = arcade.Text(
            "change_angle = False",
            3 * QUARTER_WIDTH, HALF_HEIGHT - 150,
            anchor_x='center')

        self.sprites.extend([
            self.laser_base_sprite,
            self.left_rotating_laser_sprite,
            self.platform_base_sprite,
            self.right_rotating_platform_sprite])

    def on_update(self, delta_time: float):
        # Rotate the laser sprite and change its angle
        self.left_rotating_laser_sprite.rotate_around_point(
            self.laser_base_sprite.position,
            120 * delta_time)

        # Rotate the platform sprite but don't change its angle
        self.right_rotating_platform_sprite.rotate_around_point(
            self.platform_base_sprite.position,
            60 * delta_time, False)

    def on_draw(self):
        # Draw the sprite list.
        self.clear()
        self.sprites.draw()

        self.laser_text.draw()
        self.platform_text.draw()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == '__main__':
    main()



"""
This program shows how to:
  * Display a sequence of screens in your game.  The "arcade.View"
    class makes it easy to separate the code for each screen into
    its own class.
  * This example shows the absolute basics of using "arcade.View".
    See the "different_screens_example.py" for how to handle
    screen-specific data.

Make a separate class for each view (screen) in your game.
The class will inherit from arcade.View. The structure will
look like an arcade.Window as each View will need to have its own draw,
update and window event methods. To switch a View, simply create a View
with `view = MyView()` and then use the "self.window.set_view(view)" method.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.view_screens_minimal
"""

import arcade


WIDTH = 1280
HEIGHT = 720


class MenuView(arcade.View):
    """ Class that manages the 'menu' view. """

    def on_show_view(self):
        """ Called when switching to this view"""
        self.window.background_color = arcade.color.WHITE

    def on_draw(self):
        """ Draw the menu """
        self.clear()
        arcade.draw_text("Menu Screen - click to advance", WIDTH / 2, HEIGHT / 2,
                         arcade.color.BLACK, font_size=30, anchor_x="center")

    def on_mouse_press(self, _x, _y, _button, _modifiers):
        """ Use a mouse press to advance to the 'game' view. """
        game_view = GameView()
        game_view.setup()
        self.window.show_view(game_view)


class GameView(arcade.View):
    """ Manage the 'game' view for our program. """

    def __init__(self):
        super().__init__()
        # Create variables here

    def setup(self):
        """ This should set up your game and get it ready to play """
        # Replace 'pass' with the code to set up your game
        pass

    def on_show_view(self):
        """ Called when switching to this view"""
        self.background_color = arcade.color.ORANGE_PEEL

    def on_draw(self):
        """ Draw everything for the game. """
        self.clear()
        arcade.draw_text("Game - press SPACE to advance", WIDTH / 2, HEIGHT / 2,
                         arcade.color.BLACK, font_size=30, anchor_x="center")

    def on_key_press(self, key, _modifiers):
        """ Handle key presses. In this case, we'll just count a 'space' as
        game over and advance to the game over view. """
        if key == arcade.key.SPACE:
            game_over_view = GameOverView()
            self.window.show_view(game_over_view)


class GameOverView(arcade.View):
    """ Class to manage the game over view """
    def on_show_view(self):
        """ Called when switching to this view"""
        self.window.background_color = arcade.color.BLACK

    def on_draw(self):
        """ Draw the game over view """
        self.clear()
        arcade.draw_text("Game Over - press ESCAPE to advance", WIDTH / 2, HEIGHT / 2,
                         arcade.color.WHITE, 30, anchor_x="center")

    def on_key_press(self, key, _modifiers):
        """ If user hits escape, go back to the main menu view """
        if key == arcade.key.ESCAPE:
            menu_view = MenuView()
            self.window.show_view(menu_view)


def main():
    """ Startup """
    window = arcade.Window(WIDTH, HEIGHT, "Different Views Minimal Example")
    menu_view = MenuView()
    window.show_view(menu_view)
    arcade.run()


if __name__ == "__main__":
    main()



from copy import deepcopy

from pyglet.event import EVENT_HANDLED

import arcade
from arcade import uicolor
from arcade.gui import UIEvent, UIMousePressEvent
from arcade.gui.events import UIControllerButtonPressEvent, UIOnChangeEvent, UIOnClickEvent
from arcade.gui.experimental.focus import UIFocusMixin
from arcade.gui.ui_manager import UIManager
from arcade.gui.widgets import UILayout, UIWidget
from arcade.gui.widgets.buttons import UIFlatButton
from arcade.gui.widgets.layout import UIBoxLayout


class _UIDropdownOverlay(UIFocusMixin, UIBoxLayout):
    """Represents the dropdown options overlay.

    Currently only handles closing the overlay when clicked outside of the options.
    """

    # TODO move also options logic to this class

    def show(self, manager: UIManager):
        manager.add(self, layer=UIManager.OVERLAY_LAYER)

    def hide(self):
        """Hide the overlay."""
        self.set_focus(None)
        if self.parent:
            self.parent.remove(self)

    def on_event(self, event: UIEvent) -> bool | None:
        if isinstance(event, UIMousePressEvent):
            # Click outside of dropdown options
            if not self.rect.point_in_rect((event.x, event.y)):
                self.hide()
                return EVENT_HANDLED

        if isinstance(event, UIControllerButtonPressEvent):
            # TODO find a better and more generic way to handle controller events for this
            if event.button == "b":
                self.hide()
                return EVENT_HANDLED

        return super().on_event(event)


class UIDropdown(UILayout):
    """A dropdown layout. When clicked displays a list of options provided.

    Triggers an event when an option is clicked, the event can be read by

    .. code:: py

        dropdown = Dropdown()

        @dropdown.event()
        def on_change(event: UIOnChangeEvent):
            print(event.old_value, event.new_value)

    Args:
        x: x coordinate of bottom left
        y: y coordinate of bottom left
        width: Width of each of the option.
        height: Height of each of the option.
        default: The default value shown.
        options: The options displayed when the layout is clicked.
        primary_style: The style of the primary button.
        dropdown_style: The style of the buttons in the dropdown.
        active_style: The style of the dropdown button, which represents the active option.
    """

    DIVIDER = None

    DEFAULT_BUTTON_STYLE = {
        "normal": UIFlatButton.UIStyle(
            font_color=uicolor.GREEN_NEPHRITIS,
        ),
        "hover": UIFlatButton.UIStyle(
            font_color=uicolor.WHITE,
            bg=uicolor.DARK_BLUE_WET_ASPHALT,
            border=uicolor.GRAY_CONCRETE,
        ),
        "press": UIFlatButton.UIStyle(
            font_color=uicolor.DARK_BLUE_MIDNIGHT_BLUE,
            bg=uicolor.WHITE_CLOUDS,
            border=uicolor.GRAY_CONCRETE,
        ),
        "disabled": UIFlatButton.UIStyle(
            font_color=uicolor.WHITE_SILVER,
            bg=uicolor.GRAY_ASBESTOS,
        ),
    }
    DEFAULT_DROPDOWN_STYLE = {
        "normal": UIFlatButton.UIStyle(),
        "hover": UIFlatButton.UIStyle(
            font_color=uicolor.WHITE,
            bg=uicolor.DARK_BLUE_WET_ASPHALT,
            border=uicolor.GRAY_CONCRETE,
        ),
        "press": UIFlatButton.UIStyle(
            font_color=uicolor.DARK_BLUE_MIDNIGHT_BLUE,
            bg=uicolor.WHITE_CLOUDS,
            border=uicolor.GRAY_CONCRETE,
        ),
        "disabled": UIFlatButton.UIStyle(
            font_color=uicolor.WHITE_SILVER,
            bg=uicolor.GRAY_ASBESTOS,
        ),
    }

    def __init__(
        self,
        *,
        x: float = 0,
        y: float = 0,
        width: float = 150,
        height: float = 30,
        default: str | None = None,
        options: list[str | None] | None = None,
        primary_style=None,
        dropdown_style=None,
        active_style=None,
        **kwargs,
    ):
        if primary_style is None:
            primary_style = self.DEFAULT_BUTTON_STYLE
        if dropdown_style is None:
            dropdown_style = self.DEFAULT_DROPDOWN_STYLE
        if active_style is None:
            active_style = self.DEFAULT_BUTTON_STYLE

        # TODO handle if default value not in options or options empty
        if options is None:
            options = []
        self._options = options
        self._value = default

        super().__init__(x=x, y=y, width=width, height=height, **kwargs)

        self._default_style = deepcopy(primary_style)
        self._dropdown_style = deepcopy(dropdown_style)
        self._active_style = deepcopy(active_style)

        # Setup button showing value
        self._default_button = UIFlatButton(
            text=self._value or "", width=self.width, height=self.height, style=self._default_style
        )
        self._default_button.on_click = self._on_button_click  # type: ignore

        self._overlay = _UIDropdownOverlay()
        self._update_options()

        # add children after super class setup
        self.add(self._default_button)

        self.register_event_type("on_change")

    @property
    def value(self) -> str | None:
        """Current selected option."""
        return self._value

    @value.setter
    def value(self, value: str | None):
        """Change the current selected option to a new option."""
        old_value = self._value
        self._value = value
        self._default_button.text = self._value or ""

        self._update_options()
        self.dispatch_event("on_change", UIOnChangeEvent(self, old_value, value))
        self.trigger_render()

    def _update_options(self):
        # generate options
        self._overlay.clear()

        for option in self._options:
            if option is None:  # None = UIDropdown.DIVIDER, required by pyright
                self._overlay.add(
                    UIWidget(width=self.width, height=2).with_background(color=arcade.color.GRAY)
                )
                continue
            else:
                button = self._overlay.add(
                    UIFlatButton(
                        text=option,
                        width=self.width,
                        height=self.height,
                        style=self._active_style if self.value == option else self._dropdown_style,
                    )
                )
            button.on_click = self._on_option_click

        self._overlay.detect_focusable_widgets()

    def _show_overlay(self):
        manager = self.get_ui_manager()
        if manager is None:
            raise Exception("UIDropdown could not find UIManager in its parents.")

        self._overlay.show(manager)

    def _on_button_click(self, _: UIOnClickEvent):
        self._show_overlay()

    def _on_option_click(self, event: UIOnClickEvent):
        source: UIFlatButton = event.source
        self._overlay.hide()
        self.value = source.text

    def do_layout(self):
        """Position the overlay, this is not a common thing to do in do_layout,
        but is required for the dropdown."""
        self._default_button.rect = self.rect

        # resize layout to contain widgets
        overlay = self._overlay
        rect = overlay.rect
        if overlay.size_hint_min is not None:
            rect = rect.resize(*overlay.size_hint_min)

        self._overlay.rect = rect.align_top(self.bottom - 2).align_left(self._default_button.left)

    def on_change(self, event: UIOnChangeEvent):
        """To be implemented by the user, triggered when the current selected value
        is changed to a different option.
        """
        pass



"""
A-Star Path-finding

Artwork from https://kenney.nl

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.astar_pathfinding
"""

import arcade
from arcade import camera
import random

SPRITE_IMAGE_SIZE = 128
SPRITE_SCALING = 0.25
SPRITE_SIZE = int(SPRITE_IMAGE_SIZE * SPRITE_SCALING)

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "A-Star Path-finding"

MOVEMENT_SPEED = 5

VIEWPORT_MARGIN = 100
HORIZONTAL_BOUNDARY = WINDOW_WIDTH / 2.0 - VIEWPORT_MARGIN
VERTICAL_BOUNDARY = WINDOW_HEIGHT / 2.0 - VIEWPORT_MARGIN

# If the player moves further than this boundary away from the camera we use a
# constraint to move the camera
CAMERA_BOUNDARY = arcade.LRBT(
    -HORIZONTAL_BOUNDARY,
    HORIZONTAL_BOUNDARY,
    -VERTICAL_BOUNDARY,
    VERTICAL_BOUNDARY,
)

class GameView(arcade.View):
    """
    Main application class.
    """

    def __init__(self):
        """
        Initializer
        """

        # Call the parent class initializer
        super().__init__()

        # Variables that will hold sprite lists
        self.player_list = None
        self.wall_list = None
        self.enemy_list = None

        # Set up the player info
        self.player = None

        # Track the current state of what key is pressed
        self.left_pressed = False
        self.right_pressed = False
        self.up_pressed = False
        self.down_pressed = False

        self.physics_engine = None

        # --- Related to paths
        # List of points that makes up a path between two points
        self.path = None
        # List of points we checked to see if there is a barrier there
        self.barrier_list = None

        # Set the window background color
        self.background_color = arcade.color.AMAZON

        # Camera
        self.camera = None

    def setup(self):
        """ Set up the game and initialize the variables. """

        # Sprite lists
        self.player_list = arcade.SpriteList()
        self.wall_list = arcade.SpriteList(use_spatial_hash=True,
                                           spatial_hash_cell_size=128)
        self.enemy_list = arcade.SpriteList()

        # Set up the player
        resource = ":resources:images/animated_characters/" \
                   "female_person/femalePerson_idle.png"
        self.player = arcade.Sprite(resource, scale=SPRITE_SCALING)
        self.player.center_x = SPRITE_SIZE * 5
        self.player.center_y = SPRITE_SIZE * 1
        self.player_list.append(self.player)

        # Set enemies
        resource = ":resources:images/animated_characters/zombie/zombie_idle.png"
        enemy = arcade.Sprite(resource, scale=SPRITE_SCALING)
        enemy.center_x = SPRITE_SIZE * 4
        enemy.center_y = SPRITE_SIZE * 7
        self.enemy_list.append(enemy)

        spacing = SPRITE_SIZE * 3
        for column in range(10):
            for row in range(15):
                sprite = arcade.Sprite(":resources:images/tiles/grassCenter.png",
                                       scale=SPRITE_SCALING)

                x = (column + 1) * spacing
                y = (row + 1) * sprite.height

                sprite.center_x = x
                sprite.center_y = y
                if random.randrange(100) > 30:
                    self.wall_list.append(sprite)

        self.physics_engine = arcade.PhysicsEngineSimple(self.player,
                                                         self.wall_list)

        # --- Path related
        # This variable holds the travel-path. We keep it as an attribute so
        # we can calculate it in on_update, and draw it in on_draw.
        self.path = None
        # Grid size for calculations. The smaller the grid, the longer the time
        # for calculations. Make sure the grid aligns with the sprite wall grid,
        # or some openings might be missed.
        grid_size = SPRITE_SIZE

        # Calculate the playing field size. We can't generate paths outside of
        # this.
        playing_field_left_boundary = -SPRITE_SIZE * 2
        playing_field_right_boundary = SPRITE_SIZE * 35
        playing_field_top_boundary = SPRITE_SIZE * 17
        playing_field_bottom_boundary = -SPRITE_SIZE * 2

        # This calculates a list of barriers. By calculating it here in the
        # init, we are assuming this list does not change. In this example,
        # our walls don't move, so that is ok. If we want moving barriers (such as
        # moving platforms or enemies) we need to recalculate. This can be an
        # time-intensive process depending on the playing field size and grid
        # resolution.

        # Note: If the enemy sprites are the same size, we only need to calculate
        # one of these. We do NOT need a different one for each enemy. The sprite
        # is just used for a size calculation.
        self.barrier_list = arcade.AStarBarrierList(enemy,
                                                    self.wall_list,
                                                    grid_size,
                                                    playing_field_left_boundary,
                                                    playing_field_right_boundary,
                                                    playing_field_bottom_boundary,
                                                    playing_field_top_boundary)

        self.camera = camera.Camera2D()

    def on_draw(self):
        """
        Render the screen.
        """
        # This command has to happen before we start drawing
        self.clear()

        with self.camera.activate():
            # Draw all the sprites.
            self.player_list.draw()
            self.wall_list.draw()
            self.enemy_list.draw()

            if self.path:
                arcade.draw_line_strip(self.path, arcade.color.BLUE, 2)

    def on_update(self, delta_time):
        """ Movement and game logic """

        # Calculate speed based on the keys pressed
        self.player.change_x = 0
        self.player.change_y = 0

        if self.up_pressed and not self.down_pressed:
            self.player.change_y = MOVEMENT_SPEED
        elif self.down_pressed and not self.up_pressed:
            self.player.change_y = -MOVEMENT_SPEED
        if self.left_pressed and not self.right_pressed:
            self.player.change_x = -MOVEMENT_SPEED
        elif self.right_pressed and not self.left_pressed:
            self.player.change_x = MOVEMENT_SPEED

        # Update the character
        self.physics_engine.update()

        # Calculate a path to the player
        enemy = self.enemy_list[0]
        # Set to True if we can move diagonally. Note that diagonal movement
        # might cause the enemy to clip corners.
        self.path = arcade.astar_calculate_path(enemy.position,
                                                self.player.position,
                                                self.barrier_list,
                                                diagonal_movement=False)
        # print(self.path,"->", self.player.position)

        # --- Manage Scrolling ---
        self.camera.position = camera.grips.constrain_boundary_xy(
            self.camera.view_data, CAMERA_BOUNDARY, self.player.position
        )

    def on_key_press(self, key, modifiers):
        """Called whenever a key is pressed. """

        if key in (arcade.key.UP, arcade.key.W):
            self.up_pressed = True
        elif key in (arcade.key.DOWN, arcade.key.S):
            self.down_pressed = True
        elif key in (arcade.key.LEFT, arcade.key.A):
            self.left_pressed = True
        elif key in (arcade.key.RIGHT, arcade.key.D):
            self.right_pressed = True
        # Close the window / exit game
        elif key == arcade.key.ESCAPE:
            self.window.close()

    def on_key_release(self, key, modifiers):
        """Called when the user releases a key. """

        if key in (arcade.key.UP, arcade.key.W):
            self.up_pressed = False
        elif key in (arcade.key.DOWN, arcade.key.S):
            self.down_pressed = False
        elif key in (arcade.key.LEFT, arcade.key.A):
            self.left_pressed = False
        elif key in (arcade.key.RIGHT, arcade.key.D):
            self.right_pressed = False


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""Demonstrates general setup.

If Arcade and Python are properly installed, you can run this example with:
python -m arcade.examples.gui.0_basic_setup

Content:
- create a view manually creating UIManager and adding widgets
- create a second view extending UIView and adding widgets

"""

import arcade
from arcade.gui import (
    UIManager,
    UITextureButton,
    UIAnchorLayout,
    UIView,
)

# Preload textures, because they are mostly used multiple times, so they are not
# loaded multiple times
TEX_RED_BUTTON_NORMAL = arcade.load_texture(":resources:gui_basic_assets/button/red_normal.png")
TEX_RED_BUTTON_HOVER = arcade.load_texture(":resources:gui_basic_assets/button/red_hover.png")
TEX_RED_BUTTON_PRESS = arcade.load_texture(":resources:gui_basic_assets/button/red_press.png")


class GreenView(arcade.View):
    """Uses the arcade.View and shows how to integrate UIManager."""

    def __init__(self):
        super().__init__()

        # Create a UIManager
        self.ui = UIManager()

        # Create an anchor layout, which can be used to position widgets on screen
        anchor = self.ui.add(UIAnchorLayout())

        # Add a button switch to the other View.
        button = anchor.add(
            UITextureButton(
                text="Switch to blue view",
                texture=TEX_RED_BUTTON_NORMAL,
                texture_hovered=TEX_RED_BUTTON_HOVER,
                texture_pressed=TEX_RED_BUTTON_PRESS,
            )
        )

        # add a button to switch to the blue view
        @button.event("on_click")
        def on_click(event):
            self.window.show_view(BlueView())

    def on_show_view(self) -> None:
        self.ui.enable()

    def on_hide_view(self) -> None:
        self.ui.disable()

    def on_draw(self):
        # Clear the screen
        self.clear(color=arcade.uicolor.GREEN_EMERALD)

        # Add draw commands that should be below the UI
        # ...

        self.ui.draw()

        # Add draw commands that should be on top of the UI (uncommon)
        # ...


class BlueView(UIView):
    """Uses the arcade.gui.UIView which takes care about the UIManager setup."""

    def __init__(self):
        super().__init__()
        self.background_color = arcade.uicolor.BLUE_PETER_RIVER

        # Create an anchor layout, which can be used to position widgets on screen
        anchor = self.add_widget(UIAnchorLayout())

        # Add a button switch to the other View.
        button = anchor.add(
            UITextureButton(
                text="Switch to green view",
                texture=TEX_RED_BUTTON_NORMAL,
                texture_hovered=TEX_RED_BUTTON_HOVER,
                texture_pressed=TEX_RED_BUTTON_PRESS,
            )
        )

        # add a button to switch to the green view
        @button.event("on_click")
        def on_click(event):
            self.window.show_view(GreenView())

    def on_draw_before_ui(self):
        # Add draw commands that should be below the UI
        pass

    def on_draw_after_ui(self):
        # Add draw commands that should be on top of the UI (uncommon)
        pass


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(title="GUI Example: Basic Setup")

    # Show the view on screen
    window.show_view(GreenView())

    # Start the arcade game loop
    arcade.run()

if __name__ == "__main__":
    main()



"""
Manipulate a cv2 video with shadertoy.

Dependencies:
    pip install opencv-python

"""

import cv2  # type: ignore

import arcade
from arcade.experimental.shadertoy import Shadertoy

SCREEN_WIDTH = 400
SCREEN_HEIGHT = 300
SCREEN_TITLE = "ShaderToy Video"


class ShadertoyVideo(arcade.View):
    """
    Can be used to add effects like rain to the background of the game.

    Make sure to inherit this view and call super for `__init__`,
    `on_draw`, `on_update` and `on_resize`.
    """

    def __init__(self, path: str):
        super().__init__()
        self.shadertoy = Shadertoy(
            self.window.get_framebuffer_size(),
            """
                void mainImage( out vec4 fragColor, in vec2 fragCoord )
                {
                    // Calculate the texture coordinate of the current fragment.
                    // This interpolates from 0,0 to 1,1 from lower left to upper right
                    vec2 uv = fragCoord.xy / iResolution.xy;

                    // Alter texture coordinates to make some waves
                    vec2 pos = uv - vec2(0.5);
                    float dist = length(pos) - iTime / 5.0;
                    vec2 direction = normalize(pos);
                    vec2 uv2 = uv + (direction * (sin(dist * 50.0 - iTime) - 0.5)) * 0.02;

                    fragColor = texture(iChannel0, uv2);
                }
            """,
        )
        self.video = cv2.VideoCapture(str(arcade.resources.resolve(path)))
        width, height = (
            int(self.video.get(cv2.CAP_PROP_FRAME_WIDTH)),
            int(self.video.get(cv2.CAP_PROP_FRAME_HEIGHT)),
        )
        self.video_texture = self.window.ctx.texture((width, height), components=3)
        self.video_texture.wrap_x = self.window.ctx.CLAMP_TO_EDGE
        self.video_texture.wrap_y = self.window.ctx.CLAMP_TO_EDGE
        self.video_texture.swizzle = "BGR1"
        self.shadertoy.channel_0 = self.video_texture
        self.window.set_size(width, height)

    def on_draw(self):
        self.clear()
        self.shadertoy.render()

    def on_update(self, delta_time: float):
        self.shadertoy.time += delta_time
        self.next_frame()

    def on_resize(self, width: int, height: int):
        super().on_resize(width, height)
        self.shadertoy.resize(self.window.get_framebuffer_size())

    def next_frame(self):
        exists, frame = self.video.read()
        frame = cv2.flip(frame, 0)
        if exists:
            self.video_texture.write(frame)



"""
Transforming to multiple buffers.

This examples hows how we can configure a program
to transform into multiple buffers. This is done
by specifying the varyings capture mode. By default
transforms output all data interleaved to one buffer.

Each out value in the vertex shader will end up
in a separate buffer with separate capture mode.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.gl.transform_multi
"""

import struct
import arcade

def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(1280, 720, "Transform Multi - GPU")

    num_invocations = 20
    ctx = window.ctx


    # Simple transform program outputting the vertex id
    program = ctx.program(
        vertex_shader="""
        #version 330

        out float out_a;
        out float out_b;

        void main() {
            out_a = float(gl_VertexID + 1);
            out_b = float((gl_VertexID + 1) * 2);
        }
        """,
        varyings_capture_mode="separate",
    )
    # Buffer for out_a data
    buffer_a = ctx.buffer(reserve=num_invocations * 4)
    # Buffer for out_b data
    buffer_b = ctx.buffer(reserve=num_invocations * 4)

    # Create an empty geometry instance since we don't have
    # inputs in our vertex shader
    geometry = ctx.geometry()
    # Force the vertex shader to run 10 times (vertices=10)
    geometry.transform(
        program,
        [buffer_a, buffer_b],
        vertices=num_invocations,
        # buffer_offset=4,  # We can play with offsets if needed
        # first=4,  # Test skipping vertices
    )
    # Read the buffer data from vram and print it
    print(struct.unpack(f"{num_invocations}f", buffer_a.read()))
    print(struct.unpack(f"{num_invocations}f", buffer_b.read()))


if __name__ == "__main__":
    main()



"""
Show how to have enemies shoot bullets aimed at the player.

If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.sprite_bullets_enemy_aims
"""

import arcade
import math

WINDOW_WIDTH = 1280
WINDOW_HEIGHT = 720
WINDOW_TITLE = "Sprites and Bullets Enemy Aims Example"
BULLET_SPEED = 4


class GameView(arcade.View):
    """ Main application class """

    def __init__(self):
        super().__init__()

        self.background_color = arcade.color.BLACK

        self.frame_count = 0

        self.enemy_list = None
        self.bullet_list = None
        self.player_list = None
        self.player = None

    def setup(self):
        self.enemy_list = arcade.SpriteList()
        self.bullet_list = arcade.SpriteList()
        self.player_list = arcade.SpriteList()

        # Add player ship
        self.player = arcade.Sprite(
            ":resources:images/space_shooter/playerShip1_orange.png",
            scale=0.5,
        )
        self.player_list.append(self.player)

        # Add top-left enemy ship
        enemy = arcade.Sprite(
            ":resources:images/space_shooter/playerShip1_green.png",
            scale=0.5,
        )
        enemy.center_x = 120
        enemy.center_y = WINDOW_HEIGHT - enemy.height
        enemy.angle = 180
        self.enemy_list.append(enemy)

        # Add top-right enemy ship
        enemy = arcade.Sprite(
            ":resources:images/space_shooter/playerShip1_green.png",
            scale=0.5,
        )
        enemy.center_x = WINDOW_WIDTH - 120
        enemy.center_y = WINDOW_HEIGHT - enemy.height
        enemy.angle = 180
        self.enemy_list.append(enemy)

    def on_draw(self):
        """Render the screen. """

        self.clear()

        self.enemy_list.draw()
        self.bullet_list.draw()
        self.player_list.draw()

    def on_update(self, delta_time):
        """All the logic to move, and the game logic goes here. """

        self.frame_count += 1

        # Loop through each enemy that we have
        for enemy in self.enemy_list:

            # First, calculate the angle to the player. We could do this
            # only when the bullet fires, but in this case we will rotate
            # the enemy to face the player each frame, so we'll do this
            # each frame.

            # Position the start at the enemy's current location
            start_x = enemy.center_x
            start_y = enemy.center_y

            # Get the destination location for the bullet
            dest_x = self.player.center_x
            dest_y = self.player.center_y

            # Do math to calculate how to get the bullet to the destination.
            # Calculation the angle in radians between the start points
            # and end points. This is the angle the bullet will travel.
            x_diff = dest_x - start_x
            y_diff = dest_y - start_y
            angle = -math.atan2(y_diff, x_diff) + 3.14 / 2

            # Set the enemy to face the player.
            enemy.angle = math.degrees(angle)

            # Shoot every 60 frames change of shooting each frame
            if self.frame_count % 60 == 0:
                bullet = arcade.Sprite(":resources:images/space_shooter/laserBlue01.png")
                bullet.center_x = start_x
                bullet.center_y = start_y

                # Angle the bullet sprite
                bullet.angle = math.degrees(angle) - 90

                # Taking into account the angle, calculate our change_x
                # and change_y. Velocity is how fast the bullet travels.
                bullet.change_x = math.sin(angle) * BULLET_SPEED
                bullet.change_y = math.cos(angle) * BULLET_SPEED

                self.bullet_list.append(bullet)

        # Get rid of the bullet when it flies off-screen
        for bullet in self.bullet_list:
            if bullet.top < 0:
                bullet.remove_from_sprite_lists()

        self.bullet_list.update()

    def on_mouse_motion(self, x, y, delta_x, delta_y):
        """Called whenever the mouse moves. """
        self.player.position = x, y


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(WINDOW_WIDTH, WINDOW_HEIGHT, WINDOW_TITLE)

    # Create and setup the GameView
    game = GameView()
    game.setup()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Creating compressed textures in OpenGL

Simple formats to test:
* DXT1: gl.GL_COMPRESSED_RGBA_S3TC_DXT1_EXT 6:1 compression ratio
* DXT5: gl.GL_COMPRESSED_RGBA_S3TC_DXT5_EXT 4:1 compression ratio

Running this example:
python -m arcade.examples.gl.texture_compression
"""

import PIL.Image
import arcade
import arcade.gl
from pyglet import gl


class CompressedTextures(arcade.Window):

    def __init__(self):
        super().__init__(1280, 720, "Compressed Textures Example")

        # self.texture = self.create_compressed_manual()
        self.texture = self.create_simple()

        self.geometry = arcade.gl.geometry.quad_2d_fs()
        self.program = self.ctx.program(
            vertex_shader="""
            #version 330

            in vec2 in_vert;
            in vec2 in_uv;

            out vec2 uv;

            void main() {
                gl_Position = vec4(in_vert, 0.0, 1.0);
                uv = in_uv;
            }
            """,
            fragment_shader="""
            #version 330

            uniform sampler2D tex;

            in vec2 uv;

            out vec4 fragColor;

            void main() {
                fragColor = texture(tex, uv);
            }
            """,
        )

    def create_simple(self):
        """Create a simple texture"""
        return self.ctx.load_texture(
            ":assets:images/backgrounds/abstract_1.jpg",
            internal_format=gl.GL_COMPRESSED_RGBA_S3TC_DXT1_EXT,
            compressed=True,
        )

    def create_compressed_manual(self):
        """Manually load and compress a texture"""
        path = arcade.resources.resolve(":assets:images/backgrounds/abstract_1.jpg")
        image = PIL.Image.open(path)
        components = 3 if image.mode == "RGB" else 4
        if components == 3:
            compression_format = gl.GL_COMPRESSED_RGB_S3TC_DXT1_EXT
        elif components == 4:
            compression_format = gl.GL_COMPRESSED_RGBA_S3TC_DXT5_EXT
        else:
            raise Exception(f"Invalid number of components: {components}")

        return self.ctx.texture(
            size=image.size,
            components=components,
            internal_format=compression_format,
            compressed=True,
            data=image.tobytes(),
        )

    def create_from_compressed_data(self):
        """Create data from compressed bytes"""
        # TODO: This is not easy to replicate with only arcade
        # compressed_data = <get compressed data from some source>
        # size = <the with and height of the image when decompressed>
        # components = <1, 3 or 4 depending on the image format>
        # return self.ctx.texture(
        #     size=size,
        #     components=components,
        #     internal_format=gl.GL_COMPRESSED_RGBA_S3TC_DXT5_EXT,  # tweak to match the format
        #     compressed=True,
        #     data=compressed_data,
        # )
        raise NotImplementedError

    def on_draw(self):
        self.clear()
        self.texture.use(0)
        self.geometry.render(self.program)


CompressedTextures().run()



from PIL.Image import Image

from arcade.types import Point2List

from .base import HitBox, HitBoxAlgorithm, RotatableHitBox
from .bounding_box import BoundingHitBoxAlgorithm
from .pymunk import PymunkHitBoxAlgorithm
from .simple import SimpleHitBoxAlgorithm

#: The simple hit box algorithm.
algo_simple = SimpleHitBoxAlgorithm()
#: The detailed hit box algorithm.
algo_detailed = PymunkHitBoxAlgorithm()
#: The bounding box hit box algorithm.
algo_bounding_box = BoundingHitBoxAlgorithm()
#: The default hit box algorithm.
algo_default = algo_simple


# Temporary functions for backwards compatibility
def calculate_hit_box_points_simple(image: Image, *args) -> Point2List:
    """
    Given an RGBA image, this returns points that make up a hit box around it. Attempts
    to trim out transparent pixels.

    Args:
        image: Image get hit box from.
    """
    return algo_simple.calculate(image)


def calculate_hit_box_points_detailed(
    image: Image,
    hit_box_detail: float = 4.5,
) -> Point2List:
    """
    Given an RGBA image, this returns points that make up a hit box around it. Attempts
    to trim out transparent pixels.

    Args:
        image:
            Image get hit box from.
        hit_box_detail:
            How detailed to make the hit box. There's a
            trade-off in number of points vs. accuracy.
    """
    return algo_detailed.calculate(image, detail=hit_box_detail)


__all__ = [
    "HitBoxAlgorithm",
    "HitBox",
    "RotatableHitBox",
    "SimpleHitBoxAlgorithm",
    "PymunkHitBoxAlgorithm",
    "BoundingHitBoxAlgorithm",
    "algo_simple",
    "algo_detailed",
    "algo_bounding_box",
    "algo_default",
    "calculate_hit_box_points_simple",
    "calculate_hit_box_points_detailed",
]



"""
If Python and Arcade are installed, this example can be run from the command line with:
python -m arcade.examples.music_control_demo
"""
import arcade
import arcade.gui
import arcade.gui.widgets.buttons
import arcade.gui.widgets.layout


class GameView(arcade.View):
    def __init__(self):
        super().__init__()

        self.media_player = None
        self.paused = True
        self.songs = [":resources:music/funkyrobot.mp3",
                      ":resources:music/1918.mp3"]
        self.cur_song_index = 0

        self.my_music = arcade.load_sound(self.songs[self.cur_song_index])

        # This creates a "manager" for all our UI elements
        self.ui_manager = arcade.gui.UIManager(self.window)

        box = arcade.gui.widgets.layout.UIBoxLayout(vertical=False, space_between=20)

        # --- Start button
        normal_texture = arcade.load_texture(
            ":resources:onscreen_controls/flat_dark/sound_off.png"
        )
        hover_texture = arcade.load_texture(
            ":resources:onscreen_controls/shaded_dark/sound_off.png"
        )
        press_texture = arcade.load_texture(
            ":resources:onscreen_controls/shaded_dark/sound_off.png"
        )

        # Create our button
        self.start_button = arcade.gui.widgets.buttons.UITextureButton(
            texture=normal_texture,
            texture_hovered=hover_texture,
            texture_pressed=press_texture,
        )

        # Map that button's on_click method to this view's on_button_click method.
        self.start_button.on_click = self.start_button_clicked  # type: ignore

        # Add in our element.
        box.add(self.start_button)

        # --- Down button
        press_texture = arcade.load_texture(
            ":resources:onscreen_controls/shaded_dark/down.png"
        )
        normal_texture = arcade.load_texture(
            ":resources:onscreen_controls/flat_dark/down.png"
        )
        hover_texture = arcade.load_texture(
            ":resources:onscreen_controls/shaded_dark/down.png"
        )

        # Create our button
        self.down_button = arcade.gui.widgets.buttons.UITextureButton(
            texture=normal_texture,
            texture_hovered=hover_texture,
            texture_pressed=press_texture,
        )

        # Map that button's on_click method to this view's on_button_click method.
        self.down_button.on_click = self.volume_down  # type: ignore
        self.down_button.scale(0.5)

        # Add in our element.
        box.add(self.down_button)

        # --- Up button
        press_texture = arcade.load_texture(
            ":resources:onscreen_controls/shaded_dark/up.png"
        )
        normal_texture = arcade.load_texture(
            ":resources:onscreen_controls/flat_dark/up.png"
        )
        hover_texture = arcade.load_texture(
            ":resources:onscreen_controls/shaded_dark/up.png"
        )

        # Create our button
        self.up_button = arcade.gui.widgets.buttons.UITextureButton(
            texture=normal_texture,
            texture_hovered=hover_texture,
            texture_pressed=press_texture,
        )

        # Map that button's on_click method to this view's on_button_click method.
        self.up_button.on_click = self.volume_up  # type: ignore
        self.up_button.scale(0.5)

        # Add in our element.
        box.add(self.up_button)

        # --- Right button
        press_texture = arcade.load_texture(
            ":resources:onscreen_controls/shaded_dark/right.png"
        )
        normal_texture = arcade.load_texture(
            ":resources:onscreen_controls/flat_dark/right.png"
        )
        hover_texture = arcade.load_texture(
            ":resources:onscreen_controls/shaded_dark/right.png"
        )

        # Create our button
        self.right_button = arcade.gui.widgets.buttons.UITextureButton(
            texture=normal_texture,
            texture_hovered=hover_texture,
            texture_pressed=press_texture,
        )

        # Map that button's on_click method to this view's on_button_click method.
        self.right_button.on_click = self.forward  # type: ignore
        self.right_button.scale(0.5)

        # Add in our element.
        box.add(self.right_button)

        # Place buttons in the center of the screen using
        # an UIAnchorWidget with default values
        self.ui_manager.add(
            arcade.gui.widgets.layout.UIAnchorLayout(children=[box])
        )

    def music_over(self):
        self.media_player.pop_handlers()
        self.media_player = None
        self.sound_button_off()
        self.cur_song_index += 1
        if self.cur_song_index >= len(self.songs):
            self.cur_song_index = 0
        self.my_music = arcade.load_sound(self.songs[self.cur_song_index])
        self.media_player = self.my_music.play()
        self.media_player.push_handlers(on_eos=self.music_over)

    def volume_down(self, *_):
        if self.media_player and self.media_player.volume > 0.2:
            self.media_player.volume -= 0.2

    def volume_up(self, *_):
        if self.media_player and self.media_player.volume < 1.0:
            self.media_player.volume += 0.2

    def forward(self, *_):
        skip_time = 10

        if self.media_player:
            if self.media_player.time < self.my_music.get_length() - skip_time:
                self.media_player.seek(self.media_player.time + 10)

    def sound_button_on(self):
        self.start_button.texture_pressed = \
            arcade.load_texture(":resources:onscreen_controls/shaded_dark/sound_on.png")
        self.start_button.texture = \
            arcade.load_texture(":resources:onscreen_controls/flat_dark/sound_on.png")
        self.start_button.texture_hovered = \
            arcade.load_texture(":resources:onscreen_controls/shaded_dark/sound_on.png")

    def sound_button_off(self):
        self.start_button.texture_pressed = \
            arcade.load_texture(":resources:onscreen_controls/shaded_dark/sound_off.png")
        self.start_button.texture = \
            arcade.load_texture(":resources:onscreen_controls/flat_dark/sound_off.png")
        self.start_button.texture_hovered = \
            arcade.load_texture(":resources:onscreen_controls/shaded_dark/sound_off.png")

    def start_button_clicked(self, *_):
        self.paused = False
        if not self.media_player:
            # Play button has been hit, and we need to start playing from the beginning.
            self.media_player = self.my_music.play()
            self.media_player.push_handlers(on_eos=self.music_over)
            self.sound_button_on()
        elif not self.media_player.playing:
            # Play button hit, and we need to un-pause our playing.
            self.media_player.play()
            self.sound_button_on()
        elif self.media_player.playing:
            # We are playing music, so pause.
            self.media_player.pause()
            self.sound_button_off()

    def on_draw(self):
        self.clear()

        # This draws our UI elements
        self.ui_manager.draw()
        arcade.draw_text("Music Demo",
                         x=0, y=self.window.height - 55,
                         width=self.window.width,
                         font_size=40,
                         align="center",
                         color=arcade.color.BLACK)

        if self.media_player:
            seconds = self.media_player.time
            minutes = int(seconds // 60)
            seconds = int(seconds % 60)
            arcade.draw_text(f"Time: {minutes}:{seconds:02}",
                             x=10, y=10, color=arcade.color.BLACK, font_size=24)
            volume = self.media_player.volume
            arcade.draw_text(f"Volume: {volume:3.1f}",
                             x=10, y=50, color=arcade.color.BLACK, font_size=24)

    def on_show_view(self):
        self.window.background_color = arcade.color.ALMOND

        # Registers handlers for GUI button clicks, etc.
        # We don't really use them in this example.
        self.ui_manager.enable()

    def on_hide_view(self):
        # This unregisters the manager's UI handlers,
        # Handlers respond to GUI button clicks, etc.
        self.ui_manager.disable()


def main():
    """ Main function """
    # Create a window class. This is what actually shows up on screen
    window = arcade.Window(title="Arcade Music Control Demo")

    # Create the GameView
    game = GameView()

    # Show GameView on screen
    window.show_view(game)

    # Start the arcade game loop
    arcade.run()


if __name__ == "__main__":
    main()



"""
Emitter - Invisible object that determines when Particles are emitted,
actually emits them, and manages them over their lifetime
"""

from __future__ import annotations

from collections.abc import Callable

import arcade
from arcade import Vec2
from arcade.types import Point, Velocity

from .particle import Particle


class EmitController:
    """Base class for how a client configure the rate at which an Emitter emits Particles

    Subclasses allow the client to control the rate and duration of emitting
    """

    def how_many(self, delta_time: float, current_particle_count: int) -> int:
        raise NotImplementedError("EmitterRate.how_many must be implemented")

    def is_complete(self) -> bool:
        raise NotImplementedError("EmitterRate.is_complete must be implemented")


class EmitBurst(EmitController):
    """Used to configure an Emitter to emit particles in one burst"""

    def __init__(self, count: int):
        self._is_complete = False
        self._count = count

    def how_many(self, delta_time: float, current_particle_count: int) -> int:
        if not self._is_complete:
            self._is_complete = True
            return self._count
        return 0

    def is_complete(self) -> bool:
        return True


class EmitMaintainCount(EmitController):
    """
    Used to configure an Emitter so it emits particles so that the given count
    is always maintained.
    """

    def __init__(self, particle_count: int):
        self._target_count = particle_count

    def how_many(self, delta_time: float, current_particle_count: int) -> int:
        return self._target_count - current_particle_count

    def is_complete(self) -> bool:
        return False


class EmitInterval(EmitController):
    """
    Base class used to configure an Emitter to have a constant rate of emitting.
    Will emit indefinitely.
    """

    def __init__(self, emit_interval: float):
        if emit_interval <= 0:
            raise ValueError("Invalid value for emit_interval. Must be larger than 0.")
        self._emit_interval = emit_interval
        self._carryover_time = 0.0

    def how_many(self, delta_time: float, current_particle_count: int) -> int:
        self._carryover_time += delta_time
        emit_count = 0
        while self._carryover_time >= self._emit_interval:
            self._carryover_time -= self._emit_interval
            emit_count += 1
        return emit_count

    def is_complete(self) -> bool:
        return False


class EmitterIntervalWithCount(EmitInterval):
    """
    Configure an Emitter to emit particles with given interval,
    ending after emitting given number of particles
    """

    def __init__(self, emit_interval: float, particle_count: int):
        super().__init__(emit_interval)
        self._count_remaining = particle_count

    def how_many(self, delta_time: float, current_particle_count: int) -> int:
        proposed_count = super().how_many(delta_time, current_particle_count)
        actual_count = min(proposed_count, self._count_remaining)
        self._count_remaining -= actual_count
        return actual_count

    def is_complete(self) -> bool:
        return self._count_remaining <= 0


class EmitterIntervalWithTime(EmitInterval):
    """
    Configure an Emitter to emit particles with given interval,
    ending after given number of seconds
    """

    def __init__(self, emit_interval: float, lifetime: float):
        super().__init__(emit_interval)
        self._lifetime = lifetime

    def how_many(self, delta_time: float, current_particle_count: int) -> int:
        if self._lifetime <= 0.0:
            return 0
        self._lifetime -= delta_time
        return super().how_many(delta_time, current_particle_count)

    def is_complete(self) -> bool:
        return self._lifetime <= 0


# Emitter
class Emitter:
    """
    Emits and manages Particles over their lifetime.
    The foundational class in a particle system.
    """

    def __init__(
        self,
        center_xy: Point,
        emit_controller: EmitController,
        particle_factory: Callable[[Emitter], Particle],
        change_xy: Velocity = (0.0, 0.0),
        emit_done_cb: Callable[[Emitter], None] | None = None,
        reap_cb: Callable[[], None] | None = None,
    ):
        self.change_x = change_xy[0]
        self.change_y = change_xy[1]

        self.center_x = center_xy[0]
        self.center_y = center_xy[1]
        self.angle = 0.0
        self.change_angle = 0.0
        self.rate_factory = emit_controller
        self.particle_factory = particle_factory
        self._emit_done_cb = emit_done_cb
        self._reap_cb = reap_cb
        self._particles: arcade.SpriteList[Particle] = arcade.SpriteList(use_spatial_hash=False)

    def _emit(self):
        """
        Emit one particle, its initial position and velocity are relative to the
        position and angle of the emitter.
        """
        p = self.particle_factory(self)
        p.center_x += self.center_x
        p.center_y += self.center_y

        # given the velocity, rotate it by emitter's current angle
        vel = Vec2(p.change_x, p.change_y).rotate(self.angle)

        p.change_x = vel.x
        p.change_y = vel.y
        self._particles.append(p)

    def get_count(self) -> int:
        return len(self._particles)

    def get_pos(self) -> Point:
        """Get position of emitter"""
        # TODO: should this be a property so a method call isn't needed?
        return self.center_x, self.center_y

    def update(self, delta_time: float = 1 / 60):
        # update emitter
        time_step = delta_time * 60
        self.center_x += self.change_x * time_step
        self.center_y += self.change_y * time_step
        self.angle += self.change_angle * time_step

        # update particles
        emit_count = self.rate_factory.how_many(delta_time, len(self._particles))
        for _ in range(emit_count):
            self._emit()
        self._particles.update(delta_time)
        particles_to_reap = [p for p in self._particles if p.can_reap()]
        for dead_particle in particles_to_reap:
            dead_particle.kill()

    def draw(self):
        self._particles.draw()

    def can_reap(self) -> bool:
        """Determine if Emitter can be deleted"""
        is_emit_complete = self.rate_factory.is_complete()
        can_reap = is_emit_complete and len(self._particles) <= 0
        if is_emit_complete and self._emit_done_cb:
            self._emit_done_cb(self)
            self._emit_done_cb = None
        if can_reap and self._reap_cb:
            self._reap_cb()
            self._reap_cb = None
        return can_reap



import math
from pathlib import Path
from typing import Any

import arcade
from arcade import Texture
from arcade.hitbox import HitBox, RotatableHitBox
from arcade.texture import get_default_texture
from arcade.types import PathOrTexture, Point2

from .base import BasicSprite
from .mixins import PymunkMixin

__all__ = ["Sprite"]


class Sprite(BasicSprite, PymunkMixin):
    """
    Sprites are used to render image data to the screen & perform collisions.

    Most games center around sprites. They are most frequently used as follows:

    1. Create ``Sprite`` instances from image data
    2. Add the sprites to a :py:class:`~arcade.SpriteList` instance
    3. Call :py:meth:`SpriteList.draw() <arcade.SpriteList.draw>` on the
       instance inside your ``on_draw`` method.

    For runnable examples of how to do this, please see Arcade's
    :ref:`built-in Sprite examples <sprites>`.

    .. tip:: Advanced users should see :py:class:`~arcade.BasicSprite`

        It uses fewer resources at the cost of having fewer features.

    Args:
        path_or_texture:
            Path to an image file, or a texture object.
        center_x:
            Location of the sprite in pixels.
        center_y:
            Location of the sprite in pixels.
        scale:
            Show the image at this many times its original size.
        angle:
            The initial rotation of the sprite in degrees
    """

    __slots__ = (
        "_velocity",
        "change_angle",
        "_properties",
        "boundary_left",
        "boundary_right",
        "boundary_top",
        "boundary_bottom",
        "textures",
        "cur_texture_index",
        "_hit_box",
        "physics_engines",
        "guid",
        "force",
    )

    def __init__(
        self,
        path_or_texture: PathOrTexture | None = None,
        scale: float | Point2 = 1.0,
        center_x: float = 0.0,
        center_y: float = 0.0,
        angle: float = 0.0,
        **kwargs: Any,
    ) -> None:
        if isinstance(path_or_texture, Texture):
            _texture = path_or_texture
            _textures = [_texture]
        elif isinstance(path_or_texture, str | Path):
            _texture = arcade.texture.default_texture_cache.load_or_get_texture(path_or_texture)
            _textures = [_texture]
        else:
            _texture = get_default_texture()
            # Backwards compatibility:
            # When applying default texture we don't want
            # it part of the animating ones
            _textures = []
        super().__init__(
            _texture,
            scale=scale,
            center_x=center_x,
            center_y=center_y,
            **kwargs,
        )
        PymunkMixin.__init__(self)

        self._angle = angle
        # Movement
        self._velocity = 0.0, 0.0
        self.change_angle: float = 0.0
        """Change in angle per 1/60th of a second."""

        # Custom sprite properties
        self._properties: dict[str, Any] | None = None

        # Boundaries for moving platforms in tilemaps
        self.boundary_left: float | None = None
        """
        :py:class:`~arcade.physics_engines.PhysicsEnginePlatformer`
        uses this as the left boundary for moving
        :py:attr:`~arcade.physics_engines.PhysicsEnginePlatformer.platforms`.
        """

        self.boundary_right: float | None = None
        """
        :py:class:`~arcade.physics_engines.PhysicsEnginePlatformer`
        uses this as the right boundary for moving
        :py:attr:`~arcade.physics_engines.PhysicsEnginePlatformer.platforms`.
        """

        self.boundary_top: float | None = None
        """
        :py:class:`~arcade.physics_engines.PhysicsEnginePlatformer`
        uses this as the top boundary for moving
        :py:attr:`~arcade.physics_engines.PhysicsEnginePlatformer.platforms`.
        """

        self.boundary_bottom: float | None = None
        """
        :py:class:`~arcade.physics_engines.PhysicsEnginePlatformer`
        uses this as the top boundary for moving
        :py:attr:`~arcade.physics_engines.PhysicsEnginePlatformer.platforms`.
        """

        self.cur_texture_index: int = 0
        """Current texture index for sprite animation."""
        self.textures: list[Texture] = _textures
        """List of textures stored in the sprite."""

        self.physics_engines: list[Any] = []
        """List of physics engines that have registered this sprite."""

        # Debug properties
        self.guid: str | None = None
        """A unique id for debugging purposes."""

        self._hit_box: RotatableHitBox = self._hit_box.create_rotatable(angle=self._angle)

        self._width = self._texture.width * self._scale[0]
        self._height = self._texture.height * self._scale[1]

    # --- Properties ---

    @property
    def angle(self) -> float:
        """
        Get or set the rotation or the sprite.

        The value is in degrees and is clockwise.
        """
        return self._angle

    @angle.setter
    def angle(self, new_value: float) -> None:
        if new_value == self._angle:
            return

        self._angle = new_value
        self._hit_box.angle = new_value

        for sprite_list in self.sprite_lists:
            sprite_list._update_angle(self)

        self.update_spatial_hash()

    @property
    def radians(self) -> float:
        """
        Get or set the rotation of the sprite in radians.

        The value is in radians and is clockwise.
        """
        return self._angle / 180.0 * math.pi

    @radians.setter
    def radians(self, new_value: float) -> None:
        self.angle = new_value * 180.0 / math.pi

    @property
    def velocity(self) -> Point2:
        """
        Get or set the velocity of the sprite.

        The x and y velocity can also be set separately using the
        ``sprite.change_x`` and ``sprite.change_y`` properties.

        Example::

            sprite.velocity = 1.0, 0.0
        """
        return self._velocity

    @velocity.setter
    def velocity(self, new_value: Point2) -> None:
        self._velocity = new_value

    @property
    def change_x(self) -> float:
        """Get or set the velocity in the x plane of the sprite."""
        return self.velocity[0]

    @change_x.setter
    def change_x(self, new_value: float) -> None:
        self._velocity = new_value, self._velocity[1]

    @property
    def change_y(self) -> float:
        """Get or set the velocity in the y plane of the sprite."""
        return self.velocity[1]

    @change_y.setter
    def change_y(self, new_value: float) -> None:
        self._velocity = self._velocity[0], new_value

    @property
    def hit_box(self) -> HitBox:
        """Get or set the hit box for this sprite."""
        return self._hit_box

    @hit_box.setter
    def hit_box(self, hit_box: HitBox | RotatableHitBox) -> None:
        if type(hit_box) is HitBox:
            self._hit_box = hit_box.create_rotatable(self.angle)
        else:
            # Mypy doesn't seem to understand the type check above
            # It still thinks hit_box can be a union here
            self._hit_box = hit_box  # type: ignore

    @property
    def texture(self) -> Texture:
        """Get or set the active texture for this sprite"""
        return self._texture

    @texture.setter
    def texture(self, texture: Texture) -> None:
        if texture == self._texture:
            return

        if __debug__ and not isinstance(texture, Texture):
            raise TypeError(
                f"The 'texture' parameter must be an instance of arcade.Texture,"
                f" but is an instance of '{type(texture)}'."
            )

        # If sprite is using default texture, update the hit box
        if self._texture is get_default_texture():
            self.hit_box = RotatableHitBox(
                texture.hit_box_points,
                position=self._position,
                angle=self.angle,
                scale=self._scale,
            )

        self._texture = texture
        self._width = texture.width * self._scale[0]
        self._height = texture.height * self._scale[1]
        self.update_spatial_hash()
        for sprite_list in self.sprite_lists:
            sprite_list._update_texture(self)

    @property
    def properties(self) -> dict[str, Any]:
        """Get or set custom sprite properties."""
        if self._properties is None:
            self._properties = {}
        return self._properties

    @properties.setter
    def properties(self, value: dict[str, Any]) -> None:
        self._properties = value

    # --- Movement methods -----

    def forward(self, speed: float = 1.0) -> None:
        """
        Adjusts a Sprites forward.

        Args:
            speed: The speed at which the sprite moves.
        """
        angle_rad = math.radians(self.angle)
        self.center_x += math.sin(angle_rad) * speed
        self.center_y += math.cos(angle_rad) * speed

    def reverse(self, speed: float = 1.0) -> None:
        """
        Adjusts a Sprite backwards.

        Args:
            speed: The speed at which the sprite moves.
        """
        self.forward(-speed)

    def strafe(self, speed: float = 1.0) -> None:
        """
        Adjusts a Sprite sideways.

        Args:
            speed: The speed at which the sprite moves.
        """
        angle_rad = math.radians(self.angle + 90)
        self.center_x += math.sin(angle_rad) * speed
        self.center_y += math.cos(angle_rad) * speed

    def turn_right(self, theta: float = 90.0) -> None:
        """
        Rotate the sprite right by the passed number of degrees.

        Args:
            theta: Change in angle, in degrees
        """
        self.angle = self._angle + theta

    def turn_left(self, theta: float = 90.0) -> None:
        """
        Rotate the sprite left by the passed number of degrees.

        Args:
            theta: Change in angle, in degrees
        """
        self.angle = self._angle - theta

    def stop(self) -> None:
        """
        Stop the Sprite's motion by setting the velocity and angle change to 0.
        """
        self.velocity = 0, 0
        self.change_angle = 0.0

    # ----Update Methods ----

    def update(self, delta_time: float = 1 / 60, *args, **kwargs) -> None:
        """
        The default update method for a Sprite. Can be overridden by a subclass.

        This method moves the sprite based on its velocity and angle change.

        Args:
            delta_time: Time since last update in seconds
            *args: Additional positional arguments
            **kwargs: Additional keyword arguments
        """
        # NOTE: change_x and change_y (or velocity) are historically defined as
        # the change in position per frame. To convert to change in position per
        # second, we multiply by 60 (frames per second).
        # Users can define these values in any unit they want, but this breaks
        # compatibility with physics engines. Consider changing this in the future.
        delta_time *= 60
        self.position = (
            self._position[0] + self.change_x * delta_time,
            self._position[1] + self.change_y * delta_time,
        )
        self.angle += self.change_angle * delta_time

    # ----Utility Methods----

    def update_spatial_hash(self) -> None:
        """
        Update the sprites location in the spatial hash.
        """
        # self._hit_box._adjusted_cache_dirty = True
        # super().update_spatial_hash()
        for sprite_list in self.sprite_lists:
            if sprite_list.spatial_hash is not None:
                sprite_list.spatial_hash.move(self)

    def append_texture(self, texture: Texture) -> None:
        """
        Appends a new texture to the list of textures that can be
        applied to this sprite.

        Args:
            texture: Texture to add to the list of available textures
        """
        self.textures.append(texture)

    def set_texture(self, texture_no: int) -> None:
        """
        Set the current texture by texture number.
        The number is the index into ``self.textures``.

        Args:
            texture_no: Index into ``self.textures``
        """
        texture = self.textures[texture_no]
        self.texture = texture

    def remove_from_sprite_lists(self) -> None:
        """
        Remove this sprite from all sprite lists it is in
        including registered physics engines.
        """
        super().remove_from_sprite_lists()
        for engine in self.physics_engines:
            engine.remove_sprite(self)

        self.physics_engines.clear()

    def register_physics_engine(self, physics_engine: Any) -> None:
        """
        Register a physics engine on the sprite.
        This is only needed if you actually need a reference
        to your physics engine in the sprite itself.
        It has no other purposes.

        The registered physics engines can be accessed
        through the ``physics_engines`` attribute.

        It can for example be the pymunk physics engine
        or a custom one you made.

        Args:
            physics_engine: The physics engine to register
        """
        self.physics_engines.append(physics_engine)

    def sync_hit_box_to_texture(self) -> None:
        """
        Update the sprite's hit box to match the current texture's hit box.
        """
        self.hit_box = RotatableHitBox(
            self.texture.hit_box_points,
            position=self._position,
            angle=self.angle,
            scale=self._scale,
        )
